package POE::Component::Server::Talker;

###########################################################################
### POE::Component::Server::Talker
### David Davis (xantus@cpan.org)
###
### TODO:
###  - 
###
### Copyright (c) 2004 David Davis and Teknikill software. All Rights
### Reserved. This module is free software; you can redistribute it and/or
### modify it under the same terms as Perl itself.
###########################################################################

use strict;
use warnings;

our @ISA = qw(Exporter);
our $VERSION = '0.01';

use Socket;
use Carp;
use POE qw(Session Wheel::SocketFactory Filter::Telnet
	Component::Server::Talker::ControlSession);
use MLDBM qw(SDBM_File YAML);
use Fcntl qw(:DEFAULT); # for O_CREAT & O_RDWR
use YAML qw(Dump);

our ($users_dbm,%users);

END {
	if ($users_dbm) {
		undef $users_dbm;
#		untie %users;
	}
}

sub spawn {
	my $package = shift;
	croak "$package requires an even number of parameters" if @_ % 2;
	my %params = @_;
	my $alias = $params{'Alias'};
	$alias = 'talker' unless defined($alias) and length($alias);
	$params{'Alias'} = $alias;
	$params{'ListenPort'} = $params{'ListenPort'} || 7000;
	$params{'TimeOut'} = $params{'TimeOut'} || 30;

	POE::Session->create(
#		options => { trace=>1 },
		args => [ \%params ],
		package_states => [
			'POE::Component::Server::Talker' => {
				_start				=> '_start',
				_stop				=> '_stop',
				_child				=> '_child',
				_write_log			=> '_write_log',
				register			=> 'register',
				unregister			=> 'unregister',
				notify				=> 'notify',
				accept				=> 'accept',
				tcp_error			=> 'tcp_error',
				signals				=> 'signals',
				con_notify			=> 'con_notify',
				start_listen		=> 'start_listen',

				db_store			=> 'db_store',
				
				load_modules		=> 'load_modules',
				unload_module		=> 'unload_module',
				reload_all_modules	=> 'reload_all_modules',
				reload_single_module => 'reload_single_module',
				reload_module		=> 'reload_module',
				watch_module		=> 'watch_module',
				check_module_watch	=> 'check_module_watch',
			}
		],
	);

	return 1;
}

sub _start {
	my ($kernel, $session, $heap) = @_[KERNEL, SESSION, HEAP];
	%{$heap->{params}} = %{ $_[ARG0] };

	$heap->{_main_pid} = $$;

	$session->option( @{$heap->{params}{'SessionOptions'}} ) if $heap->{params}{'SessionOptions'};
	$kernel->alias_set($heap->{params}{'Alias'});

	# watch for SIGINT
	$kernel->sig('INT', 'signals');

	my $file = "$heap->{params}->{'DataDir'}/$heap->{params}->{'UsersFile'}";
	
	# tie our database and thaw our data
	$users_dbm = tie(%users, 'MLDBM', $file, O_RDWR | O_CREAT, 0666);
	die "can't tie $file: $!" unless $users_dbm;
	
	if (!exists($users{users})) {
		$users{users} = {};
	}
	$heap->{users} = $users{users};
	
	if (!exists($users{data})) {
		$users{data} = {};
	}
	$heap->{data} = $users{data};
	
	if (!$heap->{data}{next_uid}) {
		$heap->{data}{next_uid} = 1;
	}

	$heap->{uid_sid} = { };
	
	print Dump($heap->{users},$heap->{data});
	
	$kernel->yield("load_modules");

}

sub _stop {
	my ($kernel, $heap, $session) = @_[KERNEL, HEAP, SESSION];
	$kernel->call($session->ID => _write_log => { v => 2, msg => "Server stopped at ".localtime() });
	undef $heap->{usres};
	undef $heap->{data};
}

sub _child {
	my ($kernel, $heap, $session, $action, $child) = @_[KERNEL, HEAP, SESSION, ARG0, ARG1];

	my $cheap = $child->get_heap();
	if ($cheap->{module_name}) {
		$kernel->call($session->ID => _write_log => { v => 4, msg => "Module $cheap->{module_name} $action session ".$child->ID });
#$heap->{modules}{$cheap->{module_name}} = $child->ID;
		return;
	}
	
	if ($action eq 'create') {
 		$kernel->call($session->ID => notify => 'talker_con_create' => { child_session => $child->ID });
		$kernel->call($session->ID => _write_log => { v => 4, msg => "Control session created ".$child->ID });
		$heap->{cons}{$child->ID} = { };
	} elsif ($action eq 'lose') {
 		$kernel->call($session->ID => notify => 'talker_con_destroy' => { child_session => $child->ID });
		$kernel->call($session->ID => _write_log => { v => 4, msg => "Control session lost (session ".$child->ID.")" });
		delete $heap->{cons}{$child->ID};
	}
	
	return 0;
}

sub db_store {
	my ($kernel, $session, $heap) = @_[KERNEL, SESSION, HEAP];

	# not that hard....
	$users{data} = $heap->{data};
	$users{users} = $heap->{users};
}

sub load_modules {
	my ($kernel, $session, $heap) = @_[KERNEL, SESSION, HEAP];

	if ($heap->{params}{'ModulesDir'} && -d $heap->{params}{'ModulesDir'}) {
		# add the modules dir to our 'use' path
		eval "use lib qw($heap->{params}{'ModulesDir'});";
		# look in the dir for module files
		opendir(DIR,$heap->{params}{'ModulesDir'});
		rewinddir(DIR);
		while (my $f = readdir(DIR)) {
			next unless ($f !~ m/^\./ && $f =~ m/\.pm/);
			if (-f "$heap->{params}{'ModulesDir'}/$f") {
				push(@{$heap->{module_list}},{ filename => $f, watch => 0, loaded => 0, init => 0 });
			}
		}
		closedir(DIR);
		if (ref($heap->{module_list})) {
			for my $i ( 0 .. $#{$heap->{module_list}} ) {
				my $mod = $heap->{module_list}->[$i]->{filename};
				my $fn = "$heap->{params}{'ModulesDir'}/$mod";
				$mod =~ s/\.pm$//i;
				$heap->{module_list}->[$i]->{module_name} = $mod;
				$kernel->call($session->ID => _write_log => { v => 4, msg => "Found module $mod ($fn)" });

				# use executable flag as on/off
				#unless (-x $fn) {
				#	$kernel->call($session->ID => _write_log => { v => 3, msg => "Skipping module $mod (not executable)" });
				#	next;
				#}

				# load the module
				$kernel->call($session->ID => _write_log => { v => 3, msg => "Loading module $mod" });
				eval "use $mod;";
				if ($@) {
					$kernel->call($session->ID => _write_log => { v => 2, msg => "ERROR loading module $mod : $@" });
					next;
				}
				$kernel->call($session->ID => _write_log => { v => 3, msg => "Module $mod loaded" });
				$heap->{module_list}->[$i]->{loaded} = 1;

				# init the module
				my $obj = eval { $mod->new({ params => $heap->{params}, data => $heap->{data}, users => $heap->{users} }); };
				$heap->{module_list}->[$i]->{obj} = $obj;
				if ($@) {
					$kernel->call($session->ID => _write_log => { v => 2, msg => "ERROR initializing module $mod : $@" });
					next;
				}
				$kernel->call($session->ID => _write_log => { v => 3, msg => "Module $mod initialized" });
				$heap->{module_list}->[$i]->{init} = 1;

				# collect help and regex's for commands
				unless ($obj->{provides} && ref($obj->{provides}) eq 'HASH') {
					$kernel->call($session->ID => _write_log => {
						v => 2,
						msg => "Module $mod must have a 'provides' key, and must be a hash ref, skipping it."
					});
					$heap->{module_list}->[$i]->{init} = 0;
					next;
				}

				if ($obj->{help}) {
					unless (ref($obj->{help}) eq 'ARRAY') {
						$kernel->call($session->ID => _write_log => { v => 2, msg => "Module $mod must have a 'help' key that is an array ref of scalars, skipping it." });
						$heap->{module_list}->[$i]->{init} = 0;
						next;
					}
				}
			}
			
			my $ml;
			for my $i ( 0 .. $#{$heap->{module_list}} ) {
				if ($heap->{module_list}->[$i]->{loaded} == 1 && $heap->{module_list}->[$i]->{init} == 1) {
					#$kernel->call($session->ID => _write_log => { v => 4, msg => "Module $i Ok." });
					push(@$ml, $heap->{module_list}->[$i]);
				}
			}
			$heap->{module_list} = $ml;
		} else {
			# no modules warning
		}
	}	
	
	$kernel->yield("start_listen");
}

sub reload_all_modules {
	my ($kernel, $session, $heap, $logsession) = @_[KERNEL, SESSION, HEAP, ARG0];

	# loop through our modules and fire them off in order
	for my $i ( 0 .. $#{$heap->{module_list}} ) {
		$kernel->call($session->ID => reload_module => $i => $logsession);
	}
	
}

sub reload_single_module {
	my ($kernel, $session, $heap, $module, $logsession) = @_[KERNEL, SESSION, HEAP, ARG0, ARG1];

	my $found = 0;
	# loop through our modules and fire them off in order
	for my $i ( 0 .. $#{$heap->{module_list}} ) {
		if ($heap->{module_list}->[$i]->{module_name} eq $module) {
			if ($kernel->call($session->ID => reload_module => $i => $logsession)) {
				$found = 2;
			} else {
				# found but, don't try to load a new one
				$found = 1;
			}
			last;
		}
	}

	if (!$found) {
		$kernel->call($session->ID => _write_log => { v => 4, msg => "Reload of module $module failed", s => $logsession });
		if (-f "$heap->{params}{'ModulesDir'}/$module.pm") {
			$kernel->call($session->ID => _write_log => { v => 4, msg => "Adding module $module to list", s => $logsession });
			push(@{$heap->{module_list}},{ watch => 0, init => 0, loaded => 0, filename => "$module.pm", module_name => $module });
			# love the magic -1
			unless ($kernel->call($session->ID => reload_module => -1 => $logsession)) {
				# nuke it from our list
				delete $heap->{module_list}->[-1];
			}
		}
	}	
}

sub watch_module {
	my ($kernel, $session, $heap, $module, $logsession) = @_[KERNEL, SESSION, HEAP, ARG0, ARG1];

	for my $i ( 0 .. $#{$heap->{module_list}} ) {
		if ($heap->{module_list}->[$i]->{module_name} eq $module) {
			# 1 or 0
			$heap->{module_list}->[$i]->{watch} = abs($heap->{module_list}->[$i]->{watch}-1);
			$heap->{module_list}->[$i]->{mod_time} = (stat("$heap->{params}{'ModulesDir'}/$module.pm"))[9];
			$kernel->call($session->ID => _write_log => { v => 4, msg => "Watching module $module for changes...", s => $logsession });
			$kernel->yield("check_module_watch" => $logsession);
		}
	}

}

sub check_module_watch {
	my ($kernel, $session, $heap, $logsession) = @_[KERNEL, SESSION, HEAP, ARG0];

	my $found = 0;
	for my $i ( 0 .. $#{$heap->{module_list}} ) {
		if ($heap->{module_list}->[$i]->{watch} == 1) {
			my $mod = $heap->{module_list}->[$i]->{module_name};
			if ((stat("$heap->{params}{'ModulesDir'}/$mod.pm"))[9] != $heap->{module_list}->[$i]->{mod_time}) {
				$heap->{module_list}->[$i]->{mod_time} = (stat("$heap->{params}{'ModulesDir'}/$mod.pm"))[9];
				$kernel->call($session->ID => _write_log => { v => 4, msg => "Module $mod changed!", s => $logsession });
				$kernel->call($session->ID => reload_module => $i => $logsession);
			}
			$found = 1;
		}
	}

	if ($found == 1) {
		# we stop watching if there are none to watch
		$kernel->delay_set(check_module_watch => 3 => $logsession);
	}
}

sub unload_module {
	my ($kernel, $session, $heap, $module, $logsession) = @_[KERNEL, SESSION, HEAP, ARG0, ARG1];

	for my $i ( 0 .. $#{$heap->{module_list}} ) {
		if ($heap->{module_list}->[$i]->{module_name} eq $module) {
			# shut down the module
			my $sid = $heap->{module_list}->[$i]->{obj}->{sid};
			$kernel->call($session->ID => _write_log => { v => 4, msg => "Unloading module $module ($sid)", s => $logsession });
			delete $heap->{module_list}->[$i]->{obj};
			$kernel->call($sid => 'shutdown');
			# remove it from the list
			splice(@{$heap->{module_list}},$i,1);
			
			if ($kernel->ID_id_to_session($sid)) {
				$kernel->call($session->ID => _write_log => { v => 4, msg => "Unload of module $module FAILED, session $sid is still around", s => $logsession });
			} else {
				$kernel->call($session->ID => _write_log => { v => 4, msg => "Unload of module $module succeeded", s => $logsession });
			}
			last;
		}
	}
}

sub reload_module {
	my ($kernel, $session, $heap, $i, $logsession) = @_[KERNEL, SESSION, HEAP, ARG0, ARG1];

	# this function passes s => $logsession to _write_log, which in turn sends that log
	# to the person who issued the command.  $logsession can be undef, not sending the log
	# to anyone

	# get the module name, and add .pm to it
	my $mod = $heap->{module_list}->[$i]->{module_name};
	my $modf = $mod.".pm";
	
	# is it still there?
	$kernel->call($session->ID => _write_log => { v => 4, msg => "Checking INC for $modf", s => $logsession });
	return 0 unless $INC{$modf};
	
	# shut down the module if its there
	my $sid = $heap->{module_list}->[$i]->{obj}->{sid};
	$kernel->call($session->ID => _write_log => { v => 4, msg => "Unloading module $mod (session $sid)", s => $logsession });
	delete $heap->{module_list}->[$i]->{obj};
	$kernel->call($sid => 'shutdown') if ($sid);
	
	# reload it
	$kernel->call($session->ID => _write_log => { v => 4, msg => "Reloading module $mod", s => $logsession });
	my $fh = FileHandle->new($INC{$modf});
	local($/);
	my $redef = 0;
	# counts redefined subroutines
	# I could use a postback for this...
	local($SIG{__WARN__}) = redef_count(\$redef);
	$kernel->call($session->ID => _write_log => { v => 4, msg => "Going to eval $modf", s => $logsession });
	eval <$fh>;
	if ($@) {
		$kernel->call($session->ID => _write_log => { v => 2, msg => "ERROR reloading module $mod : $@", s => $logsession });
		return 0;
	}

	# start the module
	$kernel->call($session->ID => _write_log => { v => 4, msg => "Module $mod reloaded, $redef subroutines redefined. Getting object...", s => $logsession });
	my $obj = eval { $mod->new({ params => $heap->{params}, data => $heap->{data}, users => $heap->{users} }); };
	if ($@) {
		$kernel->call($session->ID => _write_log => { v => 2, msg => "ERROR (reload) initializing module $mod : $@", s => $logsession });
		return 0;
	}

	# yay! it worked
	$heap->{module_list}->[$i]->{obj} = $obj;
	$kernel->call($session->ID => _write_log => { v => 4, msg => "Module $mod reloaded successfully", s => $logsession });

	return 1;
}

sub start_listen {
    my($kernel, $heap, $session, $sender) = @_[KERNEL, HEAP, SESSION, SENDER];

	$kernel->call($session->ID => _write_log => { v => 3, msg => "Starting SocketFactory on port $heap->{params}{'ListenPort'}" });
	
	# create a socket factory
	$heap->{wheel} = POE::Wheel::SocketFactory->new(
		BindPort       => $heap->{params}{'ListenPort'},          # on this port
		Reuse          => 'yes',          # and allow immediate port reuse
		SuccessEvent   => 'accept',       # generating this event on connection
		FailureEvent   => 'tcp_error'  # generating this event on error
	);

	$kernel->call($session->ID => _write_log => { v => 2, msg => "Listening to port $heap->{params}{ListenPort} on all interfaces." });

	# change whats shown in 'ps'
	$0 = $0." (on port $heap->{params}{ListenPort})";
}

sub con_notify {
    my($kernel, $heap, $session, $sender, $cmd, $data) = @_[KERNEL, HEAP, SESSION, SENDER, ARG0, ARG1];

	return unless ($heap->{cons});

	foreach my $h (keys %{$heap->{cons}}) {
		next if ($h == $sender->ID);
		$kernel->post($h => check_cmd => $cmd => $data);
	}
}

sub register {
    my($kernel, $heap, $session, $sender) = @_[KERNEL, HEAP, SESSION, SENDER];
    $kernel->refcount_increment($sender->ID, __PACKAGE__);
    $heap->{listeners}->{$sender->ID} = 1;
    $kernel->post($sender->ID => talker_registered => $_[SESSION]->ID);
}

sub unregister {
    my($kernel, $heap, $sender) = @_[KERNEL, HEAP, SENDER];
    $kernel->refcount_decrement($sender->ID, __PACKAGE__);
    delete $heap->{listeners}->{$sender->ID};
}

sub notify {
    my($kernel, $heap, $session, $sender, $name, $data) = @_[KERNEL, HEAP, SESSION, SENDER, ARG0, ARG1];
	# most notifys come from the control session
	# some come from a data connection and have con_session set to its parent session
	if ($sender->ID != $session->ID) {
		unless(exists($data->{con_session})) {
			$data->{con_session} = $sender->ID;
		}
	}
	
	# count up our return values and return
	# this could allow multiple sessions to deceide on things
	# ie, session1 returns 1 and session2 returns -1 (canceling out session1's return value)
	# use your imagination
	my $ret = 0;
	foreach (keys %{$heap->{listeners}}) {
    	my $tmp = $kernel->call($_ => $name => $data);
		if (defined($tmp)) {
			$ret += $tmp;
		}
	}
	
	return ($ret > 0) ? 1 : 0;
}

# Accept a new connection

sub accept {
	my ($kernel, $heap, $session, $accept_handle, $peer_addr, $peer_port) = @_[KERNEL, HEAP, SESSION, ARG0, ARG1, ARG2];

	$peer_addr = inet_ntoa($peer_addr);
	my ($port, $ip) = (sockaddr_in(getsockname($accept_handle)));
	$ip = inet_ntoa($ip);
	my $report_ip = (defined $heap->{params}{FirewallIP}) ? $heap->{params}{FirewallIP} : $ip;

	$kernel->call($session->ID => _write_log => { v => 2, msg => "Server received connection on $report_ip ($ip:$port) from $peer_addr:$peer_port" });

	# ReportIP is used to tell the client the firewall ip
	# ListenIP is the ip we actually listen on
	
	my $opt = { %{$heap->{params}} };
	$opt->{'Handle'} = $accept_handle;
	$opt->{'ListenIP'} = $ip;
	$opt->{'PeerAddr'} = $peer_addr;
	$opt->{'PeerPort'} = $peer_port;

	$opt->{'LocalIP'} = $ip;
	$opt->{'LocalPort'} = $port;
	$opt->{'ReportIP'} = $report_ip;

	unless ($kernel->call($session->ID => notify => talker_accept => {
			con_session => $session->ID,
			handle => $accept_handle,
			ssl => $_[ARG4],
			report_ip => $report_ip,
			local_ip => $ip,
			local_port => $port,
			peer_addr => $peer_addr,
			peer_port => $peer_port,
		})) {
			close($accept_handle);
			return 0;
	}
	
	$opt->{modules} = $heap->{module_list};
	$opt->{users} = $heap->{users};
	$opt->{data} = $heap->{data};
	# global user -> sid list
	$opt->{uid_sid} = $heap->{uid_sid};
	
	my $con = POE::Component::Server::Talker::ControlSession->new($opt);
	
	# keep info on the control sessions
	#$heap->{cons}{$con->{sid}} = { peer_addr => $peer_addr, peer_port => $peer_port  };
}

sub tcp_error {
	my ($kernel, $session, $operation, $errnum, $errstr) = @_[KERNEL, SESSION, ARG0, ARG1, ARG2];
	$kernel->call($session->ID => _write_log => { v => 1, msg => "Server encountered $operation error $errnum: $errstr" });
	$kernel->call($session->ID => notify => tcp_error => { con_session => $session->ID, operation => $operation, error_num => $errnum, err_str => $errstr });
}

# Handle incoming signals (INT)

sub signals {
	my ($kernel, $session, $signal_name) = @_[KERNEL, SESSION, ARG0];

	$kernel->call($session->ID => _write_log => { v => 1, msg => "Server caught SIG$signal_name" });

	# to stop ctrl-c / INT
	if ($signal_name eq 'INT') {
		#$_[KERNEL]->sig_handled();
	}

	return 0;
}

sub _write_log {
	my ($kernel, $session, $heap, $sender, $o) = @_[KERNEL, SESSION, HEAP, SENDER, ARG0];
	
	if ($o->{v} <= $heap->{params}{'LogLevel'}) {
		my $sender_sid = (defined $o->{sender_session}) ? $o->{sender_session} : ($sender->ID);
		my $type = (defined $o->{type}) ? $o->{type} : 'MAIN';

#		my $datetime = localtime();
#		print "[$datetime][$type$sender] $o->{msg}\n";

		$kernel->call($session->ID => notify => talker_write_log => {
			sender => $sender_sid,
			type => $type,
			msg => $o->{msg},
			data => $o,
		});
		
		if ($o->{s} && $o->{s} > 0) {
			$kernel->call($o->{s} => send => "LOG: $o->{msg}");
		}
	}
}

sub redef_count {
	my $ref = shift;
	sub {
		if ($_[0] =~ m/[Ss]ubroutine ([\w:]+) redefined/) {
			my $su = $1;
			$$ref++;
			local($|) = 1;
			return;
		}
		warn @_;
	};
}

1;

__END__

=head1 NAME

POE::Component::Server::Talker - Event-based Talker server

=head1 SYNOPSIS

	use POE qw(Component::Server::Talker);
	use Filesys::Virtual;

	POE::Component::Server::Talker->spawn(
		Alias           => 'talker',					# talker is default
		ListenPort      => 7000,					# port to listen on
		Version			=> 'talker '.$POE::Component::Server::Talker::VERSION,	# shown on connection
		AllowGuest		=> 1,						# 
		LogLevel		=> 4,						# 4=debug, 3=less info, 2=quiet, 1=really quiet
		TimeOut			=> 120,						# Connection Timeout
	);

	$poe_kernel->run();

=head1 DESCRIPTION

POE::Component::Server::Talker is an event driven Talker server 

=head1 AUTHOR

David Davis, E<lt>xantus@cpan.org<gt>

=head1 THANKS

Rocco Caputo, for the SSL code in POE::Component::Client::HTTP::SSL

=head1 SEE ALSO

the source code

=cut
