package POE::Component::Server::Talker::ControlSession;

###########################################################################
### POE::Component::Server::Talker::ControlSession
### David Davis (xantus@cpan.org)
###
### Copyright (c) 2004 David Davis and Teknikill software.  All Rights
### Reserved. This module is free software; you can redistribute it and/or
### modify it under the same terms as Perl itself.
###########################################################################

use strict;

use POE qw(Session Wheel::ReadWrite Driver::SysRW Filter::Telnet);
use Time::HiRes qw(time);
use FileHandle;

our %colors = (
  'reset'      => 0,
  'bold'       => 1,
  'dark'       => 2,
  'underline'  => 4,
  'underscore' => 4,
  'blink'      => 5,
  'reverse'    => 7,
  'concealed'  => 8,
  
  'black'      => 30,
  'on_black'   => 40,
  'red'        => 31,
  'on_red'     => 41,
  'green'      => 32,
  'on_green'   => 42,
  'yellow'     => 33,
  'on_yellow'  => 43,
  'blue'       => 34,
  'on_blue'    => 44,
  'magenta'    => 35,
  'on_magenta' => 45,
  'cyan'       => 36,
  'on_cyan'    => 46,
  'white'      => 37,
  'on_white'   => 47,
);

sub new {
	my $type = shift;
	my $opt = shift;

	my $self = bless { }, $type;
	
	my $ses = POE::Session->create(
		#options => { default=>1, trace=>1 },
		args => [ $opt ],
		object_states => [
			$self => {
				_start			=> '_start',
				_stop			=> '_stop',
				_default		=> '_default',
				_child			=> '_child',
				_reset_timeout	=> '_reset_timeout',
				_write_log		=> '_write_log',
				send			=> 'send',
				time_out		=> 'time_out',
				receive			=> 'receive',
				flushed			=> 'flushed',
				tcp_error		=> 'tcp_error',
				signals			=> 'signals',
				register_acct	=> 'register_acct',
				cmd_quit		=> 'cmd_quit',
				cmd_msg			=> 'cmd_msg',
				cmd_register	=> 'cmd_register',
				cmd_alias		=> 'cmd_alias',
				msg				=> 'msg',
				check_cmd		=> 'check_cmd',
				get_alias		=> 'get_alias',
				get_pass		=> 'get_pass',
				admin_cmds		=> 'admin_cmds',
			}
		],
	);

	$self->{sid} = $ses->ID;
	
	return $self;
}

sub _start {
	my ($kernel, $heap, $session, $sender, $self, $opt) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	# keep our parent session's id
	$heap->{m} = $sender->ID;

	# refs to the Talker DB
	$heap->{users} = $opt->{users};
	$heap->{data} = $opt->{data};
	# global uid -> sid list
	$heap->{uid_sid} = $opt->{uid_sid};
	
	# watch for SIGINT
	$kernel->sig('INT', 'signals');
	
	# module list
	$heap->{modules} = delete $opt->{modules};
	%{$heap->{params}} = %{ $opt };

	$heap->{input} = '';
	$heap->{cursor} = 0;
	$heap->{input_mode} = 'line';
	$heap->{ssl_mode} = 0;
	$heap->{colorize} = 1;
	$heap->{telnet} = 0;
	$heap->{telnet_sga} = 0;
	$heap->{telnet_echo} = 0;
	$heap->{telnet_eor} = 0;
	$heap->{pass_echo} = 0;
	
	if ($heap->{params}{'SslType'} == 1) {
		eval "use POE::Component::Server::Talker::SSL;";
		eval "use Symbol qw(gensym);";
		
		if ($@) {
			$kernel->call($session->ID => _write_log => 2 => "SSL/Symbol Error: $@");
			$kernel->alarm_remove_all( );
			delete $heap->{control};
			return;
		}
		
		my $old_socket = $opt->{Handle};
		my $socket = gensym();

		tie( *$socket,
			"POE::Component::Server::Talker::SSL",
			$old_socket,
			$heap->{params}{'SslPrivateKey'},
			$heap->{params}{'SslCert'},
		) or do {
			$kernel->call($session->ID => _write_log => 2 => "SSL Error: $!");
			$kernel->alarm_remove_all( );
			delete $heap->{control};
			return;
		};

		$opt->{Handle} = $socket;

		$heap->{ssl_mode} = 1;
	} elsif ($heap->{params}{'SslType'} == 2) {
		$session->register_state(cmd_secure => \&cmd_secure);
	}

	# start reading and writing
	$heap->{control} = POE::Wheel::ReadWrite->new(
		# on this handle
		Handle			=> $opt->{Handle}, 
		# using sysread and syswrite
		Driver			=> POE::Driver::SysRW->new(BlockSize => 4096), 
		Filter			=> POE::Filter::Telnet->new(),
		# generating this event for requests
		InputEvent		=> 'receive',
		# generating this event for errors
		ErrorEvent		=> 'tcp_error',
		# generating this event for all-sent
		FlushedEvent	=> 'flushed',
	);

	$heap->{auth} = 0;
	$heap->{host} = $opt->{PeerAddr};
	$heap->{port} = $opt->{PeerPort};
	
	if ($heap->{params}{'LoginTimeOut'} > 0) {
		$heap->{time_out} = $kernel->delay_set(time_out => $heap->{params}{'LoginTimeOut'});
		$kernel->call($session->ID => _write_log => 4 => "Timeout set: id ".$heap->{time_out});
	}
	
 	$kernel->call($heap->{m} => notify =>
		'talker_connected' => {
			con_session => $session->ID,
			report_ip => $opt->{ReportIP},
			local_ip => $opt->{LocalIP},
			local_port => $opt->{LocalPort},
			peer_addr => $opt->{PeerAddr},
			peer_port => $opt->{PeerPort},
		}
	);
	
	$kernel->call($session->ID => _write_log => 4 => "Control session started for $heap->{host} : $heap->{port}");

	my $greeting = eval($opt->{Greeting});
	
	if ($heap->{params}{'SslType'} == 1) {
		$kernel->delay_set(send => 1 => $greeting);
	} else {
		$kernel->yield(send => $greeting);
	}
	
	$heap->{last_prompt} = "\0[bold;green]Account\0[bold;white]:\0[reset]\0";
	$kernel->yield(send => "Please enter an account name, this differs from your alias.",
		"New users: You will set your alias later.",
		$heap->{last_prompt});
	$heap->{get_alias} = 1;
	$heap->{line_limit} = 20;
}

sub _stop {
	my ($kernel, $heap, $session) = @_[KERNEL, HEAP, SESSION];
 	$kernel->call($heap->{m} => notify => 'talker_disconnected' => { con_session => $session->ID });
	$kernel->call($session->ID => _write_log => 4 => "Control session ended with $heap->{host} : $heap->{port}");
	undef $heap->{users};
	undef $heap->{data};
	if (exists($heap->{uid}) && exists($heap->{uid_sid}->{$heap->{uid}})) {
		for my $i ( 0 .. $#{$heap->{modules}} ) {
			$kernel->call($session->ID => _write_log => 4 => "Telling module $i $heap->{modules}->[$i]->{module_name} that uid $heap->{uid} logged_out");
			$kernel->post($heap->{modules}->[$i]->{obj}->{sid} => logged_out => { uid => $heap->{uid} });
		}
		delete $heap->{uid_sid}->{$heap->{uid}};
	}
	undef $heap->{uid_sid};
}

sub _child {
	my ($kernel, $heap, $session, $action, $child) = @_[KERNEL, HEAP, SESSION, ARG0, ARG1];

	if ($action eq 'create') {
 		$kernel->call($heap->{m} => notify => 'talker_dcon_create' => { child_session => $child->ID });
		$kernel->call($session->ID => _write_log => 4 => "Child session created ".$child->ID);
		$heap->{child_session} = $child;
	} elsif ($action eq 'lose') {
 		$kernel->call($heap->{m} => notify => 'talker_dcon_destroy' => { child_session => $child->ID });
		$kernel->call($session->ID => _write_log => 4 => "Child lost (session ".$child->ID.")");
		$kernel->call($session->ID => "_reset_timeout");
		delete $heap->{child_session};
	}
	
	return 0;
}

sub send {
	my ($kernel, $session, $heap, $caller) = @_[KERNEL, SESSION, HEAP, CALLER_LINE];
	
	return unless ($heap->{control});
  	my @messages = map { (ref eq "ARRAY") ? @$_ : $_ } splice(@_,ARG0);

	unless (@messages) {
		@messages = $heap->{last_prompt};
	}
	
	foreach my $i ( 0 .. $#messages ) {
		my $txt = $messages[$i];
		if ($heap->{colorize} == 1) {
			$txt = colorize($txt,$caller);
		} else {
			$txt =~ s/\0\[[^\]]+]//g;
		}
		
		if ($txt =~ m/^\xFF/) {
			# nothing
		} elsif ($txt !~ m/\0$/ && $txt ne ""
				&& ($heap->{last_prompt} || $heap->{cursor} > 0)
				&& !$heap->{shutdown_on_flush}) {
			# print line and reprint prompt and text typed
			# get the prompt
			my $nocolor_prompt = "$heap->{last_prompt}";
			my $color_prompt = colorize($nocolor_prompt,$caller);
			# keep a uncolorized copy for measurement
			$nocolor_prompt =~ s/\0\[[^\]]+]//g;
			if ($heap->{telnet_echo} == 1) {
				if ($heap->{pass_echo} == 1) {
					$nocolor_prompt .= '*' x length($heap->{input});
				} else {
					$nocolor_prompt .= (($heap->{input} ne '') ? $heap->{input} : '');
				}
			}
			# if we are printing a shorter line than the one on the prompt line 
			my $nocolor_txt = $messages[$i];
			$nocolor_txt =~ s/\0\[[^\]]+]//g;
			$txt = "\r$txt";
			my $len1 = length($nocolor_txt);
			my $len2 = length($nocolor_prompt);
			if ($len1 < $len2) {
				my $tmp = $len2 - $len1;
				$txt .= ' ' x $tmp;
			}
			if ($i == $#messages) {
				$heap->{control}->put($txt);
				$heap->{control}->put($color_prompt.(($heap->{input} ne '') ? $heap->{input}."\0" : ''));
				return;
			}
		}
		$heap->{control}->put($txt);
	}
}

sub time_out {
	my ($kernel, $session, $heap) = @_[KERNEL, SESSION, HEAP];

	# if we have a child session, then there must be a something going on, reset the timer
	if (defined $heap->{child_session} && $heap->{params}{'LoginTimeOut'} > 0) {
		$kernel->call($heap->{m} => notify => { event => 'talker_time_out', time_inactive => $heap->{params}{'LoginTimeOut'},  });
		$heap->{time_out} = $kernel->delay_set(time_out => $heap->{params}{'LoginTimeOut'});
		$kernel->call($session->ID => _write_log => 4 => "Timeout re-set: id ".$heap->{time_out});
		return;
	}
	
	unless ($heap->{control}) {
		$kernel->alarm_remove_all();
		delete $heap->{control};
	}
	
	return if ($heap->{auth} == 1);
	
	$kernel->call($session->ID => _write_log => 2 => "Session ".$session->ID." timed out before login (".$heap->{params}{'LoginTimeOut'}.")");
	$kernel->yield(send => sprintf("\0[bold;red]You are being disconnected because you did't login before \0[bold;blue]%s\0[bold;red] seconds, Goodbye.\0[reset]",$heap->{params}{'LoginTimeOut'}));
	$kernel->alarm_remove_all();
	$heap->{shutdown_on_flush} = 1;
}

sub receive {
	my ($kernel, $session, $heap, $input) = @_[KERNEL, SESSION, HEAP, ARG0];

	if ($input =~ m/^\xFF(.*)/) {
		my @cmd = split(/:/,$1);
		$kernel->call($session->ID => _write_log => 4 => "Received telnet cmd from $heap->{host} : $heap->{port} -> $1");
		unless ($heap->{telnet} == 1) {
			$heap->{telnet} = 1;
			# IAC DO ECHO
			$kernel->yield(send => sprintf("%c%c%c",255,251,1));
			$kernel->yield(send => "Telnet client detected, \0[bold;blue]color\0[reset] is now on.");
		}	
		# telnet command received
		if ($cmd[0] eq 'TELNET_DO') {
			if ($cmd[1] eq 'TELOPT_SGA') {
				# IAC WILL SGA
				# Suppress Go Ahead
				$kernel->yield(send => sprintf("%c%c%c",255,251,3));
				$heap->{telnet_sga} = 1;
			} elsif ($cmd[1] eq 'TELOPT_ECHO') {
				# IAC WILL ECHO
				# Echo
				$kernel->yield(send => sprintf("%c%c%c",255,251,1));
				$heap->{telnet_echo} = 1;
				$heap->{input_mode} = 'char';
			} elsif ($cmd[1] eq 'TELOPT_EOR') {
				# IAC WILL EOR
				# End of record (transparent mode)
				$kernel->yield(send => sprintf("%c%c%c",255,251,25));
				$heap->{telnet_eor} = 1;
			}
		} elsif ($cmd[0] eq 'TELNET_DONT') {
			if ($cmd[1] eq 'TELOPT_SGA') {
				# IAC WONT SGA
				# Suppress Go Ahead
				$kernel->yield(send => sprintf("%c%c%c",255,252,3));
				$heap->{telnet_sga} = 0;
			} elsif ($cmd[1] eq 'TELOPT_ECHO') {
				# IAC WONT ECHO
				# Echo
				$kernel->yield(send => sprintf("%c%c%c",255,252,1));
				$heap->{telnet_echo} = 0;
			} elsif ($cmd[1] eq 'TELOPT_EOR') {
				# IAC WONT EOR
				# End of record (transparent mode)
				$kernel->yield(send => sprintf("%c%c%c",255,252,25));
				$heap->{telnet_eor} = 0;
			}
		} elsif ($cmd[0] eq 'TELNET_AYT') {
			$kernel->yield(send => "[ Yes, I'm here ]");
		} elsif ($cmd[0] eq 'TELNET_IP') {
			# quit
			$kernel->alarm_remove_all( );
			delete $heap->{control};
		} elsif ($cmd[0] eq 'TELNET_EL') {
			# erase line
			$heap->{input} = '';
			$heap->{cursor} = 0;
		} elsif ($cmd[0] eq 'TELNET_EC') {
			# backspace
			#if ($heap->{input} ne '' && $heap->{cursor} > 0) {
			#	$heap->{input} =~ s/.$//;
			#	$heap->{cursor}--;
			#}
			$input = "\x7F"; # backspace
			goto NO_RETURN;
		}
		return;
		NO_RETURN:
	}

#	$kernel->call($session->ID => _write_log => 4 => "Received input from $heap->{host} : $heap->{port} -> ".unpack('H*', $input));

#  if ($input =~ tr[\x00-\x1F][]) {
#    return;
#  }

	if (length($input) == 1) {
		my $o = ord($input);
		
		# backspace/delete	
		if ($input =~ m/\x7F|\x08/) {
			if ($heap->{input} ne '' && $heap->{cursor} > 0) {
				$heap->{input} =~ s/.$//;
				$heap->{cursor}--;
				if ($heap->{telnet_echo} == 1) {
					$kernel->yield(send => $input."\0");
				}
			}
			#$input = '';
			return;
		}
		
		if ($input eq "\n") {
			$kernel->call($session->ID => send => "");
			$input = delete $heap->{input};
			$heap->{last_input} = $input unless ($input eq '');
			$heap->{input} = '';
			$heap->{cursor} = 0;
		} else {
			if ($heap->{cursor} >= $heap->{line_limit}) {
				$input = '';
				# send bell every 2 seconds of bad entry
				if ($heap->{bell_time} && $heap->{bell_time}+2 < time()) {
					delete $heap->{bell_time};
				}
				unless ($heap->{bell_time}) {
					$heap->{bell_time} = time();
					$kernel->call($session->ID => send => "\x07\0");
				}
			} else {
				$heap->{cursor}++;
				$heap->{input} .= $input;	
				if ($heap->{telnet_echo} == 1) {
					if ($heap->{pass_echo} == 1) {
						$kernel->call($session->ID => send => "*\0");
					} else {
						$kernel->call($session->ID => send => $input."\0");
					}
				}
			}
			return;
		}
	} elsif ($input =~ s/^\x1B//) {
		# control chars, ie, left, right, up, down
		return unless ($heap->{auth} == 1);
		# ARROW UP
		# TODO: hisory instead of last_input
		if ($input eq '[A') {
			$heap->{cursor} = length($heap->{input});
			my $spaces = '';
			if (length($heap->{input}) > length($heap->{last_input})) {
				$spaces = ' ' x length($heap->{input});
			}
			$kernel->call($session->ID => send => "\r$spaces\r$heap->{last_input}\0");
			$heap->{input} = $heap->{last_input};
			$heap->{cursor} = length($heap->{input});
		} else {
			$heap->{cursor} += length($input)+1;
			$kernel->call($session->ID => send => "Input was $input cursor is now $heap->{cursor}");
			$kernel->call($session->ID => send => "^$input\0");
		}
		return;
	}
	
#    print "TEXT:".unpack('H*', $input)."\n";
	
	$input =~ s/\r/\\r/g;
	$input =~ s/\n/\\n/g;
	
	$input =~ s/.\x7f//g;
	$input =~ tr[\x00-\x1F][];

#	print "input now: [$input]\n";
  
	# TODO move this to another sub
	if ($heap->{get_alias}) {
		$kernel->yield(get_alias => $input);
		return;
	}

	if ($heap->{get_pass}) {
		$kernel->yield(get_pass => $input);
		return;
	}
		
	my $matched = 0;
	# modules get input first
	if (ref($heap->{modules}) eq 'ARRAY') {
		for my $i ( 0 .. $#{$heap->{modules}} ) {
			foreach my $event (keys %{$heap->{modules}->[$i]->{obj}->{provides}}) {
				my $regex = $heap->{modules}->[$i]->{obj}->{provides}->{$event};
				#$kernel->yield(send => "\0[bold;white]Checking $event and $regex against $input\0[reset]");
				# evals it as $input =~ m/some matching done here/
				if ((my @matches = eval "\$input =~ $regex")) {
					my $sid = $heap->{modules}->[$i]->{obj}->{sid};
					$kernel->yield(send => "\0[bold;white]Matched $event with $regex (session $sid) (".join(',',@matches).")\0[reset]");
					my $ret = $kernel->call($sid => $event => { uid => $heap->{uid}, input => $input, matches => [ @matches ] });
					if ($ret && $ret == 1) {
						$matched = 1;
						last;
					}
				}
			}
		}
	}

	# commands
	if ($input =~ m/^[\.|\!|\/]/) {
	
		if ($heap->{admin} == 1) {
			# user is an admin, process admin commands
			return if ($kernel->yield(admin_cmds => $input));
		}	
		
		# leftovers
		if (!$matched && $input =~ m/^.(\S+)(?:\s(.*))?/) {
			$kernel->post($session, "cmd_$1" => $2);
		}
	} else {
		$kernel->post($session, cmd_msg => $input);
	}
}

sub admin_cmds {
	my ($kernel, $heap, $session, $input) = @_[KERNEL, HEAP, SESSION, ARG0];

	# unescaped . matches the ! or . or /
	if ($input =~ m/^.reload(?:\s(.*))?/) {
		if ($1 && $1 eq '*') {
			$kernel->call($session->ID => send => "\0[bold;white]Reloading all modules\0[reset]");
			$kernel->call($heap->{m} => reload_all_modules => $session->ID);
		} elsif ($1) {
			$kernel->call($session->ID => send => "\0[bold;white]Reloading module $1\0[reset]");
			$kernel->call($heap->{m} => reload_single_module => $1 => $session->ID);
		} else {
			$kernel->call($session->ID => send => "\0[bold;red].reload <modulename|*>\t\tReloads a module or all modules(*)\0[reset]");
		}
		return 1;
	}
	
	if ($input =~ m/^.unload(?:\s(.*))?/) {
		if ($1) {
			$kernel->call($heap->{m} => unload_module => $1 => $session->ID);
		} else {
			$kernel->call($session->ID => send => "\0[bold;red].unload <modulename>\t\tUnloads a module\0[reset]");
		}
		return 1;
	}
	
	if ($input =~ m/^.watch(?:\s(.*))?/) {
		if ($1) {
			$kernel->call($heap->{m} => watch_module => $1 => $session->ID);
		} else {
			$kernel->call($session->ID => send => "\0[bold;red].watch <modulename>\t\tWatches a module for changes\0[reset]");
		}
		return 1;
	}
	
	if ($input =~ m/^.dump(?:\s(.*))?/) {
		if ($1) {
			my $var = $1;
			if ($var =~ s/^(\S+):://) {
				my ($found, $ret) = (0,undef);
				for my $i ( 0 .. $#{$heap->{modules}} ) {
					if ($heap->{modules}->[$i]->{module_name} eq $1) {
						$found = 1;
						$ret = $kernel->call($heap->{modules}->[$i]->{obj}->{sid} => dump => { uid => $heap->{uid} },$var);
						last;
					}
				}
				if (!$found) {
					$kernel->call($session->ID => send => "\0[bold;red]Module $1 not found\0[reset]");
					return 1;
				}
				if (!$ret) {
					$kernel->call($session->ID => send => "\0[bold;red]Module $1 does not support the dump feature\0[reset]");
				}
				return 1;
			}
			eval {
				require YAML;
			};
			if ($@) {
				$kernel->call($session->ID => send => "YAML is not installed. It must be installed to use .dump");
				return 1;
			}
			my $txt = eval "YAML::Dump($var)";
			# Id use UseHeader = 0, but dumping single vars gives me a blank line
			$txt =~ s/^--- #YAML:\d+\.\d+\s?//;
			# make it look good
			$txt =~ s/(\S+)(:\s)/\0[bold;blue]$1\0[reset]$2/g;
			$kernel->call($session->ID => send => "Dump of $var",split(/\r|\n/,$txt));
		} else {
			$kernel->call($session->ID => send => "\0[bold;red].dump <var>\t\tDumps a variable\0[reset]");
		}
		return 1;
	}
	return 0;
}

sub get_alias {
	my ($kernel, $heap, $session, $input) = @_[KERNEL, HEAP, SESSION, ARG0];

	if (length($input) > 20 || length($input) < 3) {
		$kernel->yield(send => "Account names must be \0[bold;red]3 to 20\0[reset] characters",
			"or less, that one is \0[bold;red]".length($input)."\0[reset] characters.");
		return;
	}
	
	my $acct = lc($input);
	if (exists($heap->{users}->{$acct})) {
		# existing account
		$heap->{uid} = $heap->{users}->{$acct};
		$heap->{alias} = $heap->{data}->{users}{$heap->{uid}}->{alias};
		$heap->{account} = $acct;
		$heap->{pass} = $heap->{data}->{users}{$heap->{uid}}->{pass};
		$heap->{data}->{users}{$heap->{uid}}->{logins}++;
		$kernel->yield(send => "\0[bold;blue]Welcome back, $heap->{alias}!\0[reset]");
	} else {
		# new account, set alias as 'Newbie'
		# a 'master' account must .register <account> for it to become perm.
		$heap->{alias} = "Newbie $input";
		$heap->{account} = $acct;
		$kernel->yield(send => "\0[bold;blue]Ok, now you must pick a password.\0[reset]");
	}
	delete $heap->{get_alias};
	$heap->{get_pass} = 1;
	# this is automaticly sent when a line yielded
	$heap->{last_prompt} = "\0[bold;green]Password\0[bold;white]:\0[reset]\0";
	# IAC DONT ECHO
	#$kernel->yield(send => sprintf("%c%c%c",255,254,1));
	$heap->{telnet_echo} = 0;
	#$heap->{pass_echo} = 1;
	$heap->{last_input} = '';
}

sub get_pass {
	my ($kernel, $heap, $session, $input) = @_[KERNEL, HEAP, SESSION, ARG0];

	if ($heap->{pass}) {
		unless ($input eq $heap->{pass}) {
			$heap->{tries}++;
			$kernel->yield(send => "\0[bold;red]Incorrect!\0[reset]");
			if ($heap->{tries} > 2) {
				$heap->{data}->{users}{$heap->{uid}}{failed_login} = $heap->{peer_addr};
				$kernel->call($heap->{m} => 'db_store');
				$heap->{shutdown_on_flush} = 1;
			}
			return;
		}
	} elsif (length($input) < 6) {
		$kernel->yield(send => "\0[bold;red]Sorry, your password must be at least 6 characters.\0[reset]");
		return;
	} else {
		# set password for new account
		# tell them about .alias command
		$heap->{pass} = $input;
		$kernel->yield(send => "You can change your alias with the \0[bold;yellow].alias <alias>\0[reset] command.");
	}
	delete $heap->{get_pass};
	# check global uid -> sid list
	if (exists($heap->{uid_sid}->{$heap->{uid}})) {
		# already online
		# TODO peek at session
		my $csession = $kernel->ID_id_to_session($heap->{uid_sid}->{$heap->{uid}});
		if (ref($csession) eq 'POE::Session') {
			$kernel->call($heap->{uid_sid}->{$heap->{uid}} => send => "You are being disconnected because you logged in from another location");
			$csession->get_heap()->{shutdown_on_flush} = 1;
			$heap->{last_prompt} = '';
			$kernel->yield(send => "Disconnecting your old session");
			$kernel->delay_set(get_pass => 2 => $input);
			return;
		}
	}
	# update global uid -> sid list
	$heap->{uid_sid}->{$heap->{uid}} = $session->ID;	
	$heap->{last_prompt} = '';
	$heap->{auth} = 1;
	$kernel->alarm_remove($heap->{time_out}) if (defined($heap->{time_out}));
	$kernel->yield(send => "\0[bold;cyan]Welcome!\0[reset]");
	# IAC DO ECHO
	#$kernel->yield(send => sprintf("%c%c%c",255,253,1));
	$heap->{telnet_echo} = 1;
	$heap->{line_limit} = 120;
	#$heap->{pass_echo} = 0;
	$kernel->call($heap->{m} => con_notify => send => "\0[bold;white]*$heap->{alias}*\0[reset] logged in.");
	$heap->{login_time} = time();
	$heap->{last_input} = '';

	# Temporary...
	my $ret = $kernel->call($session->ID => "register_acct");

	if ($ret == 1) {
		$kernel->yield(send => "\0[bold;yellow]You now have a registered account.\0[reset]");
	} elsif ($ret == -1) {
		$kernel->yield(send => "\0[bold;red]Someone is already using that account.\0[reset]");
	} elsif ($ret == -2) {
		# TODO add save command
		#$kernel->yield(send => "\0[bold;green]You are already registered. Maybe you want to .save?\0[reset]");
	} else {
		$kernel->yield(send => "\0[bold;red]Account registration failed.\0[reset]");
	}

	for my $i ( 0 .. $#{$heap->{modules}} ) {
		$kernel->call($session->ID => _write_log => 4 => "Telling module $i $heap->{modules}->[$i]->{module_name} that uid $heap->{uid} logged_in");
		$kernel->post($heap->{modules}->[$i]->{obj}->{sid} => logged_in => { uid => $heap->{uid} });
	}
}

sub register_acct {
	my ($kernel, $heap, $session) = @_[KERNEL, HEAP, SESSION];
	
	if (defined($heap->{uid})) {
		return -2;
	} else {
		if (exists($heap->{users}->{$heap->{account}})) {
			return -1;
		}
		$heap->{uid} = $heap->{data}{next_uid}; # get a unique id for this user
		$heap->{data}{next_uid}++; # increment for next user
		$heap->{users}->{$heap->{account}} = $heap->{uid}; # save this user_id
		$heap->{data}->{users}{$heap->{uid}} = {
			alias => $heap->{alias},
			account => $heap->{account},
			logins => 1,
			quits => 0,
			pass => $heap->{pass},
		};
		$kernel->call($heap->{m} => 'db_store');
		return 1;
	}
	
	return 0;
}

sub tcp_error {
	my ($kernel, $heap, $session, $operation, $errnum, $errstr) = @_[KERNEL, HEAP, SESSION, ARG0, ARG1, ARG2];

	if ($errnum) {
		$kernel->call($session->ID => _write_log => 3 => "Session with $heap->{host} : $heap->{port} encountered $operation error $errnum: $errstr");
	} else {
		$kernel->call($session->ID => _write_log => 3 => "Client at $heap->{host} : $heap->{port} disconnected");
	}

	$kernel->call($heap->{m} => con_notify => send => "*Thud*, $heap->{alias}'s lifeless body drops on the ground, and fades away...");
	
	$kernel->call($heap->{m} => notify => 'talker_disconnected' => { con_session => $session->ID });

	$kernel->alarm_remove_all();
	delete $heap->{control};
}

sub flushed {
	my ($kernel, $heap, $session) = @_[KERNEL, HEAP, SESSION];

	if ($heap->{shutdown_on_flush}) {
		$kernel->alarm_remove_all( );
		delete $heap->{control};
	}

	if ($heap->{start_ssl}) {
		delete $heap->{start_ssl};
		eval "use POE::Component::Server::Talker::SSL;";
		eval "use Symbol qw(gensym);";
		
		if ($@) {
			$kernel->call($session->ID => _write_log => 2 => "SSL/Symbol Error: $@");
			$kernel->alarm_remove_all( );
			delete $heap->{control};
			return;
		}
		
		$kernel->select($heap->{control}->[0]);
		
		my $old_socket = $heap->{control}->[0];
		#delete $heap->{control};
		my $socket = gensym();

		tie( *$socket,
			"POE::Component::Server::Talker::SSL",
			$old_socket,
			$heap->{params}{'SslPrivateKey'},
			$heap->{params}{'SslCert'},
		) or do {
			$kernel->call($session->ID => _write_log => 2 => "SSL Error: $!");
			$kernel->alarm_remove_all( );
			delete $heap->{control};
			return;
		};

		# start reading and writing
		$heap->{control} = POE::Wheel::ReadWrite->new(
			# on this handle
			Handle			=> $socket, 
			# using sysread and syswrite
			Driver			=> POE::Driver::SysRW->new(BlockSize => 4096), 
			Filter			=> POE::Filter::Telnet->new(),
			# generating this event for requests
			InputEvent		=> 'receive',
			# generating this event for errors
			ErrorEvent		=> 'tcp_error',
			# generating this event for all-sent
			FlushedEvent	=> 'flushed',
		);
		
		$heap->{ssl_mode} = 2;
	}
}


sub signals {
	my ($kernel, $heap, $session, $signal_name) = @_[KERNEL, HEAP, SESSION, ARG0];
	
	$kernel->call($session->ID => _write_log => 4 => "Session with $heap->{host} : $heap->{port} caught SIG $signal_name");
	# do not handle the signal
	return 0;
}

sub check_cmd {
	my ($kernel, $heap, $session, $cmd, $data) = @_[KERNEL, HEAP, SESSION, ARG0, ARG1];

	# no messages until authed
	if ($heap->{auth} == 1) {
		$kernel->yield($cmd => $data);
	}
}

# TODO keep ssl support?
sub cmd_secure {
	my ($kernel, $heap, $session, $auth) = @_[KERNEL, HEAP, SESSION, ARG0];
	
	if ($heap->{ssl_mode} == 0) {
		$auth = uc($auth);
		if ($auth eq 'SSL') {
			$heap->{start_ssl} = 1;
			$kernel->yield(send => "Starting SSL mode.");
		} else {
			$kernel->yield(send => "Security mechanism unknown. ($auth)");
		}
	} else {
		$kernel->yield(send => "Already in SSL mode.");
	}
}

sub msg {
	my ($kernel, $heap, $session, $o) = @_[KERNEL, HEAP, SESSION, ARG0];
	
	$kernel->yield(send => "\0[blue]$o->{from}\0[reset] says, \"$o->{message}\"");
}

sub cmd_alias {
	my ($kernel, $heap, $session, $alias) = @_[KERNEL, HEAP, SESSION, ARG0];

	if (!defined($alias) || $alias =~ m/^\s+$/) {
		$kernel->yield(send => "\0[blue].alias <alias>\0[reset]\t\tSet your alias; the name everyone sees.");
	} else {
		$heap->{alias} = $heap->{data}->{users}{$heap->{uid}}->{alias} = $alias;
		$kernel->call($heap->{m} => 'db_store');
		$kernel->yield(send => "\0[blue]Your alias is now \0[bold;yellow]$heap->{alias}\0[reset]");
	}
}

sub cmd_msg {
	my ($kernel, $heap, $session, $text) = @_[KERNEL, HEAP, SESSION, ARG0];
	
	$kernel->yield(msg => { message => $text, from => $heap->{alias} });
	
	$kernel->call($heap->{m} => con_notify => 'msg' => { message => $text, from => $heap->{alias} });
}

sub cmd_quit {
	my ($kernel, $heap, $session, $quit_msg) = @_[KERNEL, HEAP, SESSION];

	$quit_msg = ($quit_msg) ? " after saying \"$quit_msg\"" : '!';
	$kernel->call($heap->{m} => con_notify => send => "\0[white]In a puff of white smoke\0[reset], $heap->{alias} disappears $quit_msg");
	
	$kernel->alarm_remove_all( );
	$kernel->call($session->ID => send => "Goodbye.");
	$heap->{shutdown_on_flush} = 1;
}

sub cmd_register {
	my ($kernel, $heap, $session) = @_[KERNEL, HEAP, SESSION];

	my $ret = $kernel->call($session->ID => "register_acct");

	if ($ret == 1) {
		$kernel->yield(send => "\0[bold;yellow]You now have a registered account.\0[reset]");
	} elsif ($ret == -1) {
		$kernel->yield(send => "\0[bold;red]Someone is already using that account.\0[reset]");
	} elsif ($ret == -2) {
		$kernel->yield(send => "\0[bold;green]You are already registered. Maybe you want to .save?\0[reset]");
	} else {
		$kernel->yield(send => "\0[bold;red]Account registration failed.\0[reset]");
	}
}

sub _default {
	my ($kernel, $heap, $session, $cmd, $args) = @_[KERNEL, HEAP, SESSION, ARG0, ARG1];

	if ($cmd =~ m/^_/) {
		#$kernel->call($session->ID => _write_log => 4 => "NonHandled Event: $cmd(".join(", ", @$args).")");
	} else {
		$kernel->call($session->ID => _write_log => 4 => "Unsupported command $cmd");
		$cmd =~ s/^cmd_//;
		$kernel->yield(send => "'$cmd' is not a valid command. Type '.help commands' for a list of available commands.");
	}
	
	return 0;
}

sub _reset_timeout {
	my ($kernel,$heap) = @_[KERNEL, HEAP];
	
	if (defined $heap->{time_out}) {
		$kernel->delay_adjust( $heap->{time_out}, $heap->{params}{'LoginTimeOut'} );
	}
}

sub _write_log {
	my ($kernel, $session, $heap, $sender, $verbose, $msg) = @_[KERNEL, SESSION, HEAP, SENDER, ARG0, ARG1];
	if ($verbose <= $heap->{params}{'LogLevel'}) {
		$kernel->call($heap->{m} => _write_log => {
			type => (($sender->ID == $session->ID) ? ' CON' : ' CLD'),
			msg => $msg,
			v => $verbose,
			sender_session => $sender->ID,
		});
	}
}

my $color = sub {
	my $c = $_[0];
	my $line = $_[1];

	# should we support number colors?
	if ($c =~ m/^\d+$/ || $c =~ m/^\d+;\d+$/) {
		return "\e[${c}m";
	} else {
		if ($c =~ m/^(.+);(.*)/) {
			unless (exists($colors{$1})) {
				warn "undef color '$1' at line $line\n";
				return "";
			}
			unless (exists($colors{$2})) {
				warn "undef color '$2' at line $line\n";
				return "";
			}
			return "\e[$colors{$1};$colors{$2}m";
		} else {
			unless (exists($colors{$c})) {
				warn "undef color '$c' at line $line\n";
				return "";
			}
			return "\e[$colors{$c}m";
		}
	}
};

sub colorize {
  my ($msg, $line) = @_;

  $msg =~ s/\0\[([^\]]+)\]/$color->($1,$line)/eg;

  return $msg;
}
1;

