package POE::Filter::Telnet;

use strict;

use vars qw($VERSION);
$VERSION = do {my@r=(q$Revision: 0.01 $=~/\d+/g);sprintf"%d."."%04d"x$#r,@r};

use Carp qw(carp croak);

sub DEBUG () { 0 }

sub FRAMING_BUFFER   () { 0 }
sub INPUT_REGEXP     () { 1 }
sub OUTPUT_LITERAL   () { 2 }
sub AUTODETECT_STATE () { 3 }
sub EOR				 () { 4 }
sub SGA				 () { 5 }
sub CHAR_MODE		 () { 6 }

sub AUTO_STATE_DONE   () { 0x00 }
sub AUTO_STATE_FIRST  () { 0x01 }
sub AUTO_STATE_SECOND () { 0x02 }

our %telnet_cmds = (
	255 => 'TELNET_IAC',
	254 => 'TELNET_DONT',
	253 => 'TELNET_DO',
	252 => 'TELNET_WONT',
	251 => 'TELNET_WILL',
	250 => 'TELNET_SB',
	249 => 'TELNET_GA',
	248 => 'TELNET_EL',
	247 => 'TELNET_EC',
	246 => 'TELNET_AYT',
	245 => 'TELNET_AO',
	244 => 'TELNET_IP',
	243 => 'TELNET_BREAK',
	242 => 'TELNET_DM',
	241 => 'TELNET_NOP',
	240 => 'TELNET_SE',
	239 => 'TELNET_EOR',
	238 => 'TELNET_ABORT',
	237 => 'TELNET_SUSP',
	236 => 'TELNET_EOF',
	242 => 'TELNET_SYNCH',
);

our %telnet_opts = (
	0 => 'TELOPT_BINARY',
	1 => 'TELOPT_ECHO',
	2 => 'TELOPT_RCP',
	3 => 'TELOPT_SGA',
	4 => 'TELOPT_NAMS',
	5 => 'TELOPT_STATUS',
	6 => 'TELOPT_TM',
	7 => 'TELOPT_RCTE',
	8 => 'TELOPT_NAOL',
	9 => 'TELOPT_NAOP',
	10 => 'TELOPT_NAOCRD',
	11 => 'TELOPT_NAOHTS',
	12 => 'TELOPT_NAOHTD',
	13 => 'TELOPT_NAOFFD',
	14 => 'TELOPT_NAOVTS',
	15 => 'TELOPT_NAOVTD',
	16 => 'TELOPT_NAOLFD',
	17 => 'TELOPT_XASCII',
	18 => 'TELOPT_LOGOUT',
	19 => 'TELOPT_BM',
	20 => 'TELOPT_DET',
	21 => 'TELOPT_SUPDUP',
	22 => 'TELOPT_SUPDUPOUTPUT',
	23 => 'TELOPT_SNDLOC',
	24 => 'TELOPT_TTYPE',
	25 => 'TELOPT_EOR',
	26 => 'TELOPT_TUID',
	27 => 'TELOPT_OUTMRK',
	28 => 'TELOPT_TTYLOC',
	29 => 'TELOPT_3270REGIME',
	30 => 'TELOPT_X3PAD',
	31 => 'TELOPT_NAWS',
	32 => 'TELOPT_TSPEED',
	33 => 'TELOPT_LFLOW',
	34 => 'TELOPT_LINEMODE',
	35 => 'TELOPT_XDISPLOC',
	36 => 'TELOPT_OLD_ENVIRON',
	37 => 'TELOPT_AUTHENTICATION',
	38 => 'TELOPT_ENCRYPT',
	39 => 'TELOPT_NEW_ENVIRON',
	255 => 'TELOPT_EXOPL',
);

sub TELNET_IAC ()	    {255}; # interpret as command:
sub TELNET_DONT	()	    {254}; # you are not to use option
sub TELNET_DO ()	    {253}; # please, you use option
sub TELNET_WONT ()	    {252}; # I won't use option
sub TELNET_WILL ()	    {251}; # I will use option
sub TELNET_SB ()	    {250}; # interpret as subnegotiation
sub TELNET_GA ()	    {249}; # you may reverse the line
sub TELNET_EL ()	    {248}; # erase the current line
sub TELNET_EC ()	    {247}; # erase the current character
sub TELNET_AYT ()	    {246}; # are you there
sub TELNET_AO ()	    {245}; # abort output--but let prog finish
sub TELNET_IP ()	    {244}; # interrupt process--permanently
sub TELNET_BREAK ()	    {243}; # break
sub TELNET_DM ()	    {242}; # data mark--for connect. cleaning
sub TELNET_NOP ()	    {241}; # nop
sub TELNET_SE ()	    {240}; # end sub negotiation
sub TELNET_EOR ()	    {239}; # end of record (transparent mode)
sub TELNET_ABORT ()	    {238}; # Abort process
sub TELNET_SUSP ()	    {237}; # Suspend process
sub TELNET_EOF ()	    {236}; # End of file
sub TELNET_SYNCH ()	    {242}; # for telfunc calls

sub TELOPT_BINARY ()	      {0}; # Binary Transmission
sub TELOPT_ECHO ()	      {1}; # Echo
sub TELOPT_RCP ()	      {2}; # Reconnection
sub TELOPT_SGA ()	      {3}; # Suppress Go Ahead
sub TELOPT_NAMS ()	      {4}; # Approx Message Size Negotiation
sub TELOPT_STATUS ()	      {5}; # Status
sub TELOPT_TM ()	      {6}; # Timing Mark
sub TELOPT_RCTE ()	      {7}; # Remote Controlled Trans and Echo
sub TELOPT_NAOL ()	      {8}; # Output Line Width
sub TELOPT_NAOP ()	      {9}; # Output Page Size
sub TELOPT_NAOCRD ()	     {10}; # Output Carriage-Return Disposition
sub TELOPT_NAOHTS ()	     {11}; # Output Horizontal Tab Stops
sub TELOPT_NAOHTD ()	     {12}; # Output Horizontal Tab Disposition
sub TELOPT_NAOFFD ()	     {13}; # Output Formfeed Disposition
sub TELOPT_NAOVTS ()	     {14}; # Output Vertical Tabstops
sub TELOPT_NAOVTD ()	     {15}; # Output Vertical Tab Disposition
sub TELOPT_NAOLFD ()	     {16}; # Output Linefeed Disposition
sub TELOPT_XASCII ()	     {17}; # Extended ASCII
sub TELOPT_LOGOUT ()	     {18}; # Logout
sub TELOPT_BM ()	     {19}; # Byte Macro
sub TELOPT_DET ()	     {20}; # Data Entry Terminal
sub TELOPT_SUPDUP ()	     {21}; # SUPDUP
sub TELOPT_SUPDUPOUTPUT ()   {22}; # SUPDUP Output
sub TELOPT_SNDLOC ()	     {23}; # Send Location
sub TELOPT_TTYPE ()	     {24}; # Terminal Type
sub TELOPT_EOR ()	     {25}; # End of Record
sub TELOPT_TUID ()	     {26}; # TACACS User Identification
sub TELOPT_OUTMRK ()	     {27}; # Output Marking
sub TELOPT_TTYLOC ()	     {28}; # Terminal Location Number
sub TELOPT_3270REGIME ()     {29}; # Telnet 3270 Regime
sub TELOPT_X3PAD ()	     {30}; # X.3 PAD
sub TELOPT_NAWS ()	     {31}; # Negotiate About Window Size
sub TELOPT_TSPEED ()	     {32}; # Terminal Speed
sub TELOPT_LFLOW ()	     {33}; # Remote Flow Control
sub TELOPT_LINEMODE ()	     {34}; # Linemode
sub TELOPT_XDISPLOC ()	     {35}; # X Display Location
sub TELOPT_OLD_ENVIRON ()    {36}; # Environment Option
sub TELOPT_AUTHENTICATION () {37}; # Authentication Option
sub TELOPT_ENCRYPT ()	     {38}; # Encryption Option
sub TELOPT_NEW_ENVIRON ()    {39}; # New Environment Option
sub TELOPT_EXOPL ()	    {255}; # Extended-Options-List

#------------------------------------------------------------------------------

sub new {
  my $type = shift;

  croak "$type requires an even number of parameters" if @_ and @_ & 1;
  my %params = @_;

  croak "$type cannot have both Regexp and Literal line endings"
    if defined $params{Regexp} and defined $params{Literal};

  my ($input_regexp, $output_literal);
  my $autodetect = AUTO_STATE_DONE;

  # Literal newline for both incoming and outgoing.  Every other known
  # parameter conflicts with this one.
  if (defined $params{Literal}) {
    croak "A defined Literal must have a nonzero length"
      unless defined($params{Literal}) and length($params{Literal});
    $input_regexp   = quotemeta $params{Literal};
    $output_literal = $params{Literal};
    croak "$type cannot have Literal with any other parameter"
      if ( exists $params{InputLiteral} or # undef means something
           defined $params{InputRegexp} or
           defined $params{OutputLiteral}
         );
  }

  # Input and output are specified separately, then.
  else {

    # Input can be either a literal or a regexp.  The regexp may be
    # compiled or not; we don't rightly care at this point.
    if (exists $params{InputLiteral}) {
      $input_regexp = $params{InputLiteral};

      # InputLiteral is defined.  Turn it into a regexp and be done.
      # Otherwise we will autodetect it.
      if (defined($input_regexp) and length($input_regexp)) {
        $input_regexp = quotemeta $input_regexp;
      }
      else {
        $autodetect   = AUTO_STATE_FIRST;
        $input_regexp = '';
      }

      croak "$type cannot have both InputLiteral and InputRegexp"
        if defined $params{InputRegexp};
    }
    elsif (defined $params{InputRegexp}) {
      $input_regexp = $params{InputRegexp};
      croak "$type cannot have both InputLiteral and InputRegexp"
        if defined $params{InputLiteral};
    }
    else {
      $input_regexp = "(\\x0D\\x0A?|\\x0A\\x0D?)";
    }

    if (defined $params{OutputLiteral}) {
      $output_literal = $params{OutputLiteral};
    }
    else {
      $output_literal = "\x0D\x0A";
    }
  }

  delete @params{qw(Literal InputLiteral OutputLiteral InputRegexp)};
  carp("$type ignores unknown parameters: ", join(', ', sort keys %params))
    if scalar keys %params;

  my $self =
    bless [ '',              # FRAMING_BUFFER
            $input_regexp,   # INPUT_REGEXP
            $output_literal, # OUTPUT_LITERAL
            $autodetect,     # AUTODETECT_STATE
			'0',			 # EOR
			'0',			 # SGA
			'0',			 # CHAR_MODE
          ], $type;
		  
  DEBUG and warn join ':', @$self;

  $self;
}

#------------------------------------------------------------------------------

sub get {
  my ($self, $stream) = @_;
  my @lines;

  DEBUG and do {
    my $temp = join '', @$stream;
    $temp = unpack 'H*', $temp;
    warn "char: $self->[CHAR_MODE] got some raw data: $temp\n";
  };

  $self->[FRAMING_BUFFER] .= join '', @$stream;

  # Process as many newlines an we can find.
LINE:
  while (1) {
	
	if ($self->[FRAMING_BUFFER] =~ s/^\xFF(.)//) {
		my $opt1 = ord($1);
		if ($opt1 < TELNET_IAC && $opt1 > TELNET_SB) {
			# second command after TELNET_DONT(254),TELNET_DO(253),TELNET_WONT(252),TELNET_WILL(251)
			if ($self->[FRAMING_BUFFER] =~ s/^(.)//) {
				my $opt2 = ord($1);
				DEBUG and warn "recd telnet cmd: $opt1=".$telnet_cmds{$opt1}." $opt2=".$telnet_opts{$opt2};
				push(@lines,sprintf("%c%s:%s",255,$telnet_cmds{$opt1},$telnet_opts{$opt2}));
				next LINE;
			} else {
				# Put it back for later
				$self->[FRAMING_BUFFER] = sprintf("%c%c",255,$opt1).$self->[FRAMING_BUFFER];
			}
		} else {
			DEBUG and warn "recd telnet cmd: $opt1=".$telnet_cmds{$opt1};
			push(@lines,sprintf("%c%s",255,$telnet_cmds{$opt1}));
			next LINE;
		}
	}

	if ($self->[CHAR_MODE] == 1) {
		if ($self->[FRAMING_BUFFER] =~ s/^(.)//) {
			my $tmp = $1;
			if ($tmp =~ m/\x0D|\0/) {
				$self->[INPUT_REGEXP] = '(\x0D\x00)';
				$self->[FRAMING_BUFFER] = $tmp.$self->[FRAMING_BUFFER];
			} else {
				push(@lines,$tmp);
				next LINE;
			}
		}
	}
			
    # Autodetect is done, or it never started.  Parse some buffer!
    unless ($self->[AUTODETECT_STATE]) {
      DEBUG and warn unpack 'H*', $self->[INPUT_REGEXP];
      last LINE
        unless $self->[FRAMING_BUFFER] =~ s/^(.*?)$self->[INPUT_REGEXP]//s;
      DEBUG and warn "got line: <<", unpack('H*', $1), ">>\n";
	  if ($self->[CHAR_MODE] == 1) {
  	    push @lines, "\n";
		} else {
	      push @lines, $1;
		}
      next LINE;
    }

    # Waiting for the first line ending.  Look for a generic newline.
    if ($self->[AUTODETECT_STATE] & AUTO_STATE_FIRST) {
      last LINE
        unless $self->[FRAMING_BUFFER] =~ s/^(.*?)(\x0D\x0A?|\x0A\x0D?)//;
	  if ($self->[CHAR_MODE] == 1) {
  	    push @lines, "\n";
		} else {
	      push @lines, $1;
		}

      # The newline can be complete under two conditions.  First: If
      # it's two characters.  Second: If there's more data in the
      # framing buffer.  Loop around in case there are more lines.
      if ( (length($2) == 2) or
           (length $self->[FRAMING_BUFFER])
         ) {
        DEBUG and warn "detected complete newline after line: <<$1>>\n";
        $self->[INPUT_REGEXP] = $2;
        $self->[AUTODETECT_STATE] = AUTO_STATE_DONE;
        next LINE;
      }

      # The regexp has matched a potential partial newline.  Save it,
      # and move to the next state.  There is no more data in the
      # framing buffer, so we're done.
      DEBUG and warn "detected suspicious newline after line: <<$1>>\n";
      $self->[INPUT_REGEXP] = $2;
      $self->[AUTODETECT_STATE] = AUTO_STATE_SECOND;
      last LINE;
    }

    # Waiting for the second line beginning.  Bail out if we don't
    # have anything in the framing buffer.
    if ($self->[AUTODETECT_STATE] & AUTO_STATE_SECOND) {
      last LINE unless length $self->[FRAMING_BUFFER];

      # Test the first character to see if it completes the previous
      # potentially partial newline.
      if ( substr($self->[FRAMING_BUFFER], 0, 1) eq
           ( $self->[INPUT_REGEXP] eq "\x0D" ? "\x0A" : "\x0D" )
         ) {

        # Combine the first character with the previous newline, and
        # discard the newline from the buffer.  This is two statements
        # for backward compatibility.
        DEBUG and warn "completed newline after line: <<$1>>\n";
        $self->[INPUT_REGEXP] .= substr($self->[FRAMING_BUFFER], 0, 1);
        substr($self->[FRAMING_BUFFER], 0, 1) = '';
      }
      elsif (DEBUG) {
        warn "decided prior suspicious newline is okay\n";
      }

      # Regardless, whatever is in INPUT_REGEXP is now a complete
      # newline.  End autodetection, post-process the found newline,
      # and loop to see if there are other lines in the buffer.
      $self->[INPUT_REGEXP] = $self->[INPUT_REGEXP];
      $self->[AUTODETECT_STATE] = AUTO_STATE_DONE;
      next LINE;
    }

    die "consistency error: AUTODETECT_STATE = $self->[AUTODETECT_STATE]";
  }

  \@lines;
}

#------------------------------------------------------------------------------
# 2001-07-27 RCC: Add get_one_start() and get_one() to correct filter
# changing and make input flow control possible.

sub get_one_start {
  my ($self, $stream) = @_;

  DEBUG and do {
    my $temp = join '', @$stream;
    $temp = unpack 'H*', $temp;
    warn "char: $self->[CHAR_MODE] got some raw data: $temp\n";
  };

  $self->[FRAMING_BUFFER] .= join '', @$stream;
}

# -><- There is a lot of code duplicated here.  What can be done?

sub get_one {
  my $self = shift;

  # Process as many newlines an we can find.
LINE:
  while (1) {

	if ($self->[FRAMING_BUFFER] =~ s/^\xFF(.)//) {
		my $opt1 = ord($1);
		if ($opt1 < TELNET_IAC && $opt1 > TELNET_SB) {
			# second command after TELNET_DONT(254),TELNET_DO(253),TELNET_WONT(252),TELNET_WILL(251)
			if ($self->[FRAMING_BUFFER] =~ s/^(.)//) {
				my $opt2 = ord($1);
				DEBUG and warn "recd telnet cmd: $opt1=".$telnet_cmds{$opt1}." $opt2=".$telnet_opts{$opt2};
				return [ sprintf("%c%s:%s",255,$telnet_cmds{$opt1},$telnet_opts{$opt2}) ];
			} else {
				# Put it back for later
				$self->[FRAMING_BUFFER] = sprintf("%c%c",255,$opt1).$self->[FRAMING_BUFFER];
			}
		} else {
			DEBUG and warn "recd telnet cmd: $opt1=".$telnet_cmds{$opt1};
			return [ sprintf("%c%s",255,$telnet_cmds{$opt1}) ];
		}
	}

	if ($self->[FRAMING_BUFFER] =~ s/^(\x1B..)//) {
		return [ $1 ];
	}
	
	if ($self->[CHAR_MODE] == 1) {
		if ($self->[FRAMING_BUFFER] =~ s/^(.)//) {
			my $tmp = $1;
			if ($tmp =~ m/\x0D|\0/) {
				$self->[INPUT_REGEXP] = '(\x0D\x00)';
				$self->[FRAMING_BUFFER] = $tmp.$self->[FRAMING_BUFFER];
			} elsif ($tmp =~ m/\x1B/) {
				$self->[FRAMING_BUFFER] = $tmp.$self->[FRAMING_BUFFER];
				last;
			} else {
				return [ $tmp ];
			}
		}
	}
	
    # Autodetect is done, or it never started.  Parse some buffer!
    unless ($self->[AUTODETECT_STATE]) {
#      DEBUG and warn $self->[INPUT_REGEXP];
      last LINE unless $self->[FRAMING_BUFFER] =~ s/^(.*?)$self->[INPUT_REGEXP]//s;
      DEBUG and warn "got line: <<", unpack('H*', $1), ">>\n";

	  if ($self->[CHAR_MODE] == 1) {
		if ($1) {
			my $tmp = $1;
			$tmp =~ s/\x0D//;
			return [ $tmp ];
		} else {
			return [ "\n" ];
		}
		} else {
     	 return [ $1 ];
		}
    }

    # Waiting for the first line ending.  Look for a generic newline.
    if ($self->[AUTODETECT_STATE] & AUTO_STATE_FIRST) {
      last LINE
        unless $self->[FRAMING_BUFFER] =~ s/^(.*?)(\x0D\x0A?|\x0A\x0D?)//;

      my $line = $1;

      # The newline can be complete under two conditions.  First: If
      # it's two characters.  Second: If there's more data in the
      # framing buffer.  Loop around in case there are more lines.
      if ( (length($2) == 2) or
           (length $self->[FRAMING_BUFFER])
         ) {
        DEBUG and warn "detected complete newline after line: <<$1>>\n";
        $self->[INPUT_REGEXP] = $2;
        $self->[AUTODETECT_STATE] = AUTO_STATE_DONE;
      }

      # The regexp has matched a potential partial newline.  Save it,
      # and move to the next state.  There is no more data in the
      # framing buffer, so we're done.
      else {
        DEBUG and warn "detected suspicious newline after line: <<$1>>\n";
        $self->[INPUT_REGEXP] = $2;
        $self->[AUTODETECT_STATE] = AUTO_STATE_SECOND;
      }

	  if ($self->[CHAR_MODE] == 1) {
		return [ "\n" ];
		} else {
     	 return [ $line ];
		}
    }

    # Waiting for the second line beginning.  Bail out if we don't
    # have anything in the framing buffer.
    if ($self->[AUTODETECT_STATE] & AUTO_STATE_SECOND) {
      return [ ] unless length $self->[FRAMING_BUFFER];

      # Test the first character to see if it completes the previous
      # potentially partial newline.
      if ( substr($self->[FRAMING_BUFFER], 0, 1) eq
           ( $self->[INPUT_REGEXP] eq "\x0D" ? "\x0A" : "\x0D" )
         ) {

        # Combine the first character with the previous newline, and
        # discard the newline from the buffer.  This is two statements
        # for backward compatibility.
        DEBUG and warn "completed newline after line: <<$1>>\n";
        $self->[INPUT_REGEXP] .= substr($self->[FRAMING_BUFFER], 0, 1);
        substr($self->[FRAMING_BUFFER], 0, 1) = '';
      }
      elsif (DEBUG) {
        warn "decided prior suspicious newline is okay\n";
      }

      # Regardless, whatever is in INPUT_REGEXP is now a complete
      # newline.  End autodetection, post-process the found newline,
      # and loop to see if there are other lines in the buffer.
	  #$self->[INPUT_REGEXP] = $self->[INPUT_REGEXP];
      $self->[AUTODETECT_STATE] = AUTO_STATE_DONE;
      next LINE;
    }

    die "consistency error: AUTODETECT_STATE = $self->[AUTODETECT_STATE]";
  }

  return [ ];
}

#------------------------------------------------------------------------------
# New behavior.  First translate system newlines ("\n") into whichever
# newlines are supposed to be sent.  Second, add a trailing newline if
# one doesn't already exist.  Since the referenced output list is
# supposed to contain one line per element, we also do a split and
# join.  Bleah. ... why isn't the code doing what the comment says?

sub put {
  my ($self, $lines) = @_;

  my @raw;
  foreach (@$lines) {
	if (m/^\xFF(.{2})?/) {
        DEBUG and warn "Sending Telnet cmd\n";
		if (defined $1) {
			my ($opt1, $opt2) = split(//,$1);
			my ($opt1, $opt2) = (ord($opt1), ord($opt2));
        	DEBUG and warn "cmd $opt1 and $opt2\n";
			if ($opt1 == TELNET_WILL && $opt2 == TELOPT_ECHO) {
				$self->[CHAR_MODE] = 1;
				#$self->[AUTODETECT_STATE] = AUTO_STATE_DONE;
        		DEBUG and warn "CHAR mode ON\n";
			} elsif ($opt1 == TELNET_WONT && $opt2 == TELOPT_ECHO) {
				$self->[CHAR_MODE] = 0;
        		DEBUG and warn "CHAR mode OFF\n";
			} elsif ($opt1 == TELNET_WILL && $opt2 == TELOPT_SGA) {
				$self->[SGA] = 1;
			} elsif ($opt1 == TELNET_WONT && $opt2 == TELOPT_SGA) {
				$self->[SGA] = 0;
			} elsif ($opt1 == TELNET_WILL && $opt2 == TELOPT_EOR) {
				$self->[EOR] = 1;
			} elsif ($opt1 == TELNET_WONT && $opt2 == TELOPT_EOR) {
				$self->[EOR] = 0;
			}
		}
    	push @raw, $_;
	} elsif (s/\0$//) {
	    push @raw, $_;
	} else {
	    push @raw, $_ . $self->[OUTPUT_LITERAL];
	}
  }
  if (@raw) {
	if ($self->[EOR] == 1) {
		push @raw, sprintf("%c%c", TELNET_IAC, TELNET_EOR);
	} elsif ($self->[SGA] == 0) {
		push @raw, sprintf("%c%c", TELNET_IAC, TELNET_GA);
	}
  }

  \@raw;
}

#------------------------------------------------------------------------------

sub get_pending {
  my $self = shift;
  return [ $self->[FRAMING_BUFFER] ] if length $self->[FRAMING_BUFFER];
  return undef;
}

###############################################################################
1;

__END__

=head1 NAME

POE::Filter::Telnet - filter data as lines, and filters telnet codes

=head1 SYNOPSIS

  $filter = POE::Filter::Telnet->new();
  $arrayref_of_lines =
    $filter->get($arrayref_of_raw_chunks_from_driver);
  $arrayref_of_streamable_chunks_for_driver =
    $filter->put($arrayref_of_lines);
  $arrayref_of_leftovers =
    $filter->get_pending();

  # Use a literal newline terminator for input and output:
  $filter = POE::Filter::Telnet->new( Literal => "\x0D\x0A" );

  # Terminate input lines with a string regexp:
  $filter = POE::Filter::Telnet->new( InputRegexp   => '[!:]',
                                    OutputLiteral => "!"
                                  );

  # Terminate input lines with a compiled regexp (requires perl 5.005
  # or newer):
  $filter = POE::Filter::Telnet->new( InputRegexp   => qr/[!:]/,
                                    OutputLiteral => "!"
                                  );

  # Autodetect the input line terminator:
  $filter = POE::Filter::Telnet->new( InputLiteral => undef );

=head1 DESCRIPTION

The Telnet filter translates streams to and from separated lines.  The
lines it returns do not include the line separator (usually newlines).
Neither should the lines given to it.

Incoming newlines are recognized with a simple regular expression by
default: C</(\x0D\x0A?|\x0A\x0D?)/>.  This regexp encompasses all the
variations of CR and/or LF, but it has a race condition.

Consider a CRLF newline is broken into two stream chunks, one which
ends with CR and the other which begins with LF:

   some stream dataCR
   LFother stream data

The default regexp will recognize the CR as one end-of-line marker and
the LF as another.  The line filter will emit two lines: "some stream
data" and a blank line.  B<People are advised to specify custom
literal newlines or autodetect the newline style in applications where
blank lines are significant.>

Outgoing lines have traditional network newlines (CRLF) appended to
them by default.

=head1 PUBLIC FILTER METHODS

Please see POE::Filter.

=head1 SEE ALSO

POE::Filter.

The SEE ALSO section in L<POE> contains a table of contents covering
the entire POE distribution.

=head1 BUGS

The default input newline regexp has a race condition where incomplete
newlines can generate spurious blank input lines.

=head1 AUTHORS & COPYRIGHTS

Please see L<POE> for more information about authors and contributors.

=cut
