package rooms;

use POE;
use MLDBM qw(SDBM_File YAML);
use Fcntl qw(:DEFAULT); # for O_CREAT & O_RDWR
use YAML qw(Dump);
use Time::HiRes qw(time);

our ($rooms_dbm,%rooms);

our %op = (
	n => 's',
	s => 'n',
	e => 'w',
	w => 'e',
	ne => 'sw',
	sw => 'ne',
	nw => 'se',
	se => 'nw',
	u => 'd',
	d => 'u',
);

END {
	if ($rooms_dbm) {
		undef $rooms_dbm;
#		untie %rooms;
	}
}

sub new {
	my $class = shift;

	# every module must have a 'provides' arrayhash ref
	# and a help arrayref in its obj
	my $self = bless {
		# used to match events to commands, keep the # down to very few
		# this is matched against every line that starts with ! or . or /
		# m/^[\!|\.|\/]/
		provides => {
			# n e w s
			go_direction	=> 'm/^[\.|\/]([n|e|w|s|ne|nw|se|sw|d|u])$/i',
			connect_room	=> 'm/^[\.|\/]room connect ([n|e|w|s|ne|nw|se|sw|d|u]) (\d+)$/i',
			connect_help	=> 'm/^[\.|\/]room connect$/i',
			disconnect_room	=> 'm/^[\.|\/]room disconnect ([n|e|w|s|ne|nw|se|sw|d|u])$/i',
			disconnect_help	=> 'm/^[\.|\/]room disconnect$/i',
			room_info		=> 'm/^[\.|\/]room info(?:\s(\d+))?/i',
			room_list		=> 'm/^[\.|\/]room list$/i',
			say				=> 'm/^![\.|\/|\!](.*)/',
		},
		# used to fill the help list (optional, but very helpful :))
		# format is '.command' or '!command' or '/command'
		#help => [
		#	'.rooms',
		#],
	},$class;

	my $session = POE::Session->create(
		heap => {
			module_name => 'Rooms',
		},
		#options => { default=>1, trace=>1 },
		args => [ @_ ],
		object_states => [
			$self => {
				_start			=> '_start',
				_stop			=> '_stop',
				shutdown		=> 'shutdown',
				load_rooms		=> 'load_rooms',
				# debug event
				dump			=> 'dump',

				go_direction	=> 'go_direction',
				connect_room	=> 'connect_room',
				connect_help	=> 'help',
				disconnect_room => 'disconnect_room',
				disconnect_help	=> 'help',
				room_info		=> 'room_info',
				room_list		=> 'room_list',
				say				=> 'say',
				
				# provided for us
				logged_in		=> 'logged_in',
				logged_out		=> 'logged_out',
			}
		],
	);

	$self->{sid} = $session->ID;

	return $self;
}


sub _start {
	my ($kernel, $heap, $session, $sender, $self, $opt) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	$heap->{m} = $sender->ID;

	# users and global data
	$heap->{users} = $opt->{users};
	$heap->{data} = $opt->{data};
	$heap->{params} = $opt->{params};
	
	# keep us alive
    $kernel->refcount_increment($session->ID, "module::".__PACKAGE__);
	
	my $dir = "$heap->{params}->{'DataDir'}/rooms";
	
	if (!-d $dir) {
		# TODO create or die
	}
	
	# tie our database and thaw our data
	$rooms_dbm = tie(%rooms, 'MLDBM', "$dir/rooms.db", O_RDWR | O_CREAT, 0666);
	die "can't tie $dir/rooms.db: $!" unless $rooms_dbm;

	if (!exists($rooms{loc})) {
		$rooms{loc} = {};
	}
	$heap->{loc} = $rooms{loc};
	
	if (!exists($rooms{rdata})) {
		$rooms{rdata} = {};
	}
	$heap->{rdata} = $rooms{rdata};
	
	if (!$heap->{rdata}{next_lid}) {
		$heap->{rdata}{next_lid} = 1;
	}
	
	print Dump($heap->{rooms},$heap->{rdata});

	$kernel->yield("load_rooms");
}

sub _stop {
	my ($kernel, $heap, $session, $sender, $self) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT];

	undef $heap->{data};
	undef $heap->{users};
}

sub shutdown {
	my ($kernel, $heap, $session, $sender, $self) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT];

	for my $i ( 0 .. $#{$heap->{rooms}} ) {
		$kernel->call($heap->{rooms}->[$i]->{obj}->{sid} => 'shutdown');
	}
	
	$kernel->alarm_remove_all();
    $kernel->refcount_decrement($session->ID, "module::".__PACKAGE__);

	return $session->ID;
}

sub load_rooms {
	my ($kernel, $session, $heap) = @_[KERNEL, SESSION, HEAP];

	my $dir = "$heap->{params}{'DataDir'}/rooms";
	if ($heap->{params}{'DataDir'} && -d $dir) {
		# add the rooms dir to our 'use' path
		eval "use lib qw($dir);";
		# look in the dir for room files
		opendir(DIR,"$dir");
		rewinddir(DIR);
		while (my $f = readdir(DIR)) {
			next unless ($f !~ m/^\./ && $f =~ m/\.pm/);
			if (-f "$dir/$f") {
				push(@{$heap->{rooms}},{ filename => $f, watch => 0, init => 0 });
			}
		}
		closedir(DIR);
		if (ref($heap->{rooms})) {
			for my $i ( 0 .. $#{$heap->{rooms}} ) {
				my $mod = $heap->{rooms}->[$i]->{filename};
				my $modf = $mod;
				my $fn = "$dir/$mod";
				$mod =~ s/\.pm$//i;
				$kernel->call($heap->{m} => _write_log => { v => 4, msg => "Found room $mod ($fn)" });

				# This handles room reloads on this module being reloaded
				if (exists($INC{$modf})) {
					$kernel->call($heap->{m} => _write_log => { v => 4, msg => "Room $mod was already in memory, reloading ($modf)" });
					my $fh = FileHandle->new($INC{$modf});
					local($/);
					my $redef = 0;
					# counts redefined subroutines
					# I could use a postback for this...
					local($SIG{__WARN__}) = redef_count(\$redef);
					$kernel->call($heap->{m} => _write_log => { v => 4, msg => "Going to eval $modf" });
					eval <$fh>;
					if ($@) {
						$kernel->call($heap->{m} => _write_log => { v => 2, msg => "ERROR reloading room $mod ($modf): $@" });
						return 0;
					}
				}
				
				# use executable flag as on/off
				unless (-x $fn) {
					$kernel->call($heap->{m} => _write_log => { v => 3, msg => "Skipping room $mod (not executable)" });
					next;
				}

				# load the room
				$kernel->call($heap->{m} => _write_log => { v => 3, msg => "Loading room $mod" });
				eval "use $mod;";
				if ($@) {
					$kernel->call($heap->{m} => _write_log => { v => 2, msg => "ERROR loading room $mod : $@" });
					next;
				}
				$kernel->call($heap->{m} => _write_log => { v => 3, msg => "Room $mod loaded" });

				# init the room
				my $obj = eval { $mod->new({ params => $heap->{params}, m => $heap->{m}, users => $heap->{users}, data => $heap->{data}, rdata => $heap->{rdata}, loc => $heap->{loc} });  };
				$heap->{rooms}->[$i]->{obj} = $obj;
				if ($@) {
					$kernel->call($heap->{m} => _write_log => { v => 2, msg => "ERROR initializing room $mod : $@" });
					next;
				}
				$kernel->call($heap->{m} => _write_log => { v => 3, msg => "Room $mod initialized and created session $obj->{sid}" });
				$heap->{rooms}->[$i]->{init} = 1;
				
				if (!exists($heap->{loc}{$mod})) {
					# generate an id for the room
					$heap->{loc}{$mod} = $heap->{rdata}{next_lid};
					$heap->{rdata}{next_lid}++;
				}
				
				$kernel->call($heap->{m} => _write_log => { v => 4, msg => "$mod is room # $heap->{loc}{$mod} ($obj->{room_name})" });
				
				$heap->{loc_i}{$heap->{loc}{$mod}} = $i;
				$heap->{i_loc}{$i} = $heap->{loc}{$mod};
				$heap->{loc_sid}{$heap->{loc}{$mod}} = $obj->{sid};

				if ($obj->{entrance} && $obj->{entrance} == 1) {
					$kernel->call($heap->{m} => _write_log => { v => 4, msg => "Rooms entry point located, room # $heap->{loc}{$mod} ($obj->{room_name})" });
					# found entry point
					$heap->{rdata}{entrance_room} = $heap->{loc}{$mod};
				}
			}
			
			#my $rl;
			#for my $i ( 0 .. $#{$heap->{rooms}} ) {
			#	if ($heap->{rooms}->[$i]->{loaded} == 1 && $heap->{rooms}->[$i]->{init} == 1) {
			#		#$kernel->call($heap->{m} => _write_log => { v => 4, msg => "Room $i Ok." });
			#		push(@$rl, $heap->{rooms}->[$i]);
			#	}
			#}
			#$heap->{rooms} = $rl;
			
			# save the room info
			$rooms{loc} = $heap->{loc};
			$rooms{rdata} = $heap->{rdata};
		} else {
			# no modules warning
		}
	}
}

sub say {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	my $room1 = $heap->{rdata}{user_loc}{$o->{uid}};
	# loop through uids in the current room
	foreach my $u (keys %{$heap->{rdata}{rooms}{$room1}{users}}) {
		# tell everyone the msg
		$kernel->post($heap->{user_sids}{$u} => send => sprintf("%s says \"%s\"",$heap->{data}->{users}{$o->{uid}}{alias},$o->{input}));
	}
	
	return 1;
}

sub help {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];
	
	if ($o->{input} =~ m/room connect/i) {
		$kernel->post($sender->ID => send => ".room connect <direction> <room#>");
	} elsif ($o->{input} =~ m/room disconnect/i) {
		$kernel->post($sender->ID => send => ".room disconnect <direction>");
	} else {
		$kernel->post($sender->ID => send => "Help is unavailable for that command");
	}

	return 1;
}

sub dump {
	my ($kernel, $heap, $session, $sender, $self, $o, $var) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0, ARG1];
	require YAML;
	my $txt = eval "YAML::Dump($var)";
	# Id use UseHeader = 0, but dumping single vars gives me a blank line
	$txt =~ s/^--- #YAML:\d+\.\d+\s?//;
	# make it look good
	$txt =~ s/(\S+)(:\s)/\0[bold;blue]$1\0[reset]$2/g;
	$kernel->call($sender->ID => 'send' => "Dump of $var",split(/\r|\n/,$txt));
	return 1;
}

sub connect_room {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	my ($room1, $room2, $direction) = ($heap->{rdata}{user_loc}{$o->{uid}}, $o->{matches}->[1], $o->{matches}->[0]);
	$direction = lc($direction);
	
	if (exists($heap->{rdata}{map}{$room1}) && exists($heap->{rdata}{map}{$room1}{$direction})) {
		$kernel->post($sender->ID => send => sprintf("Room %s is connected to room %s at direction %s",
			$room1,$heap->{rdata}{map}{$room1}{$direction},$direction));
	} else {
		$heap->{rdata}{map}{$room1}{$direction} = $room2;
		# TODO connect both ways?
		#$heap->{rdata}{map}{$room2}{$op{$direction}} = $room1;
		$rooms{rdata} = $heap->{rdata};
		$kernel->post($sender->ID => send => sprintf("Room %s is NOW connected to room %s at direction %s",
			$room1,$room2,$direction));
	}
	
	return 1;
}

sub disconnect_room {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	my ($room1, $direction) = ($heap->{rdata}{user_loc}{$o->{uid}}, $o->{matches}->[0]);

	$direction = lc($direction);

	if (exists($heap->{rdata}{map}{$room1}) && exists($heap->{rdata}{map}{$room1}{$direction})) {
		$kernel->post($sender->ID => send => sprintf("Room %s is NOW disconnected from room %s at direction %s",
			$room1,$heap->{rdata}{map}{$room1}{$direction},$direction));
		delete $heap->{rdata}{map}{$room1}{$direction};
		$rooms{rdata} = $heap->{rdata};
	} else {
		$kernel->post($sender->ID => send => sprintf("Room %s doesn't have an exit at direction %s",
			$room1,$direction));
	}
	
	return 1;
}

sub room_list {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	my @txt = "Rooms:";
	foreach my $i ( 0 .. $#{$heap->{rooms}} ) {
		push(@txt,sprintf("Room # %s (%s)",$heap->{i_loc}{$i},$heap->{rooms}->[$i]->{obj}->{room_name}));
	}
	
	$kernel->post($sender->ID => send => @txt);
	
	return 1;
}

sub room_info {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	my $room1 = ($o->{matches}->[0]) ? $o->{matches}->[0] : $heap->{rdata}{user_loc}{$o->{uid}};
	unless (exists($heap->{rooms}->[$heap->{loc_i}{$room1}])) {
		$kernel->post($sender->ID => send => "Room $room1 does not exist");
		return;
	}
	my @txt = sprintf("Room # \0[bold;green]%s\0[reset] (\0[bold;blue]%s\0[reset])",
		$room1,$heap->{rooms}->[$heap->{loc_i}{$room1}]->{obj}->{room_name});

	if ($heap->{rdata}{map}{$room1}) {
		foreach my $d (sort keys %{$heap->{rdata}{map}{$room1}}) {
			push(@txt,sprintf("%s is connected to room # %s (%s)",$d,
				$heap->{rdata}{map}{$room1}{$d},
				$heap->{rooms}->[$heap->{loc_i}{$heap->{rdata}{map}{$room1}{$d}}]->{obj}->{room_name}));
		}
	} else {
		push(@txt,"There are no exits");
	}
	
	foreach my $u (keys %{$heap->{rdata}{rooms}{$room1}{users}}) {
		push(@txt,sprintf("%s (%s) is there",$u,$heap->{data}->{users}{$u}{alias}));
	}
	
	$kernel->post($sender->ID => send => @txt);
	
	return 1;
}

sub logged_in {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	my $room;
	if (exists($heap->{rdata}{user_loc}{$o->{uid}})) {
		$room = $heap->{rdata}{user_loc}{$o->{uid}};
	} elsif ($heap->{rdata}{entrance_room}) {
		$room = $heap->{rdata}{entrance_room};
	} else {
		# TODO limbo?
		return;
	}

	# keep this so we can map uid's to sids
	$heap->{user_sids}{$o->{uid}} = $sender->ID;
	
	my $sid = $heap->{loc_sid}{$room};
	$kernel->call($heap->{m} => _write_log => { v => 4, msg => sprintf("uid %s walked in, putting in room %s, session %s",
		$o->{uid},$room,$sid) });
	# rooms room# users uid = time walked in
	$heap->{rdata}{rooms}{$room}{users}{$o->{uid}} = time();
	# users uid = room#
	$heap->{rdata}{user_loc}{$o->{uid}} = $room;
	
	$rooms{rdata} = $heap->{rdata};
	#$kernel->post($heap->{rooms}->[$heap->{loc_i}{$room}]->obj->{sid} => walked_in => $o-{uid});
	$kernel->post($sid => walked_in => $sender->ID => $o);
	
	return 1;
}

sub logged_out {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	$kernel->post($heap->{rdata}{user_loc}{$o->{uid}} => walked_out => $o);
	delete $heap->{user_sids}{$o->{uid}};
}

sub go_direction {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	my $direction = lc($o->{matches}->[0]);
	#$kernel->post($sender->ID => send => "Ok, going $direction");
	
	my $room1 = $heap->{rdata}{user_loc}{$o->{uid}};
	if (exists($heap->{rdata}{map}{$room1}) && exists($heap->{rdata}{map}{$room1}{$direction})) {
		my $room2 = $heap->{rdata}{map}{$room1}{$direction};
		$kernel->post($heap->{loc_sid}{$room1} => walked_out => $sender->ID => { direction => $direction, uid => $o->{uid} });
		# take them out of the first room
		delete $heap->{rdata}{rooms}{$room1}{users}{$o->{uid}};
		# put them in the next room
		$heap->{rdata}{user_loc}{$o->{uid}} = $room2;
		$heap->{rdata}{rooms}{$room2}{users}{$o->{uid}} = time();
		# send walking in from opposite direction
		$kernel->post($heap->{loc_sid}{$room2} => walked_in => $sender->ID => { direction => $op{$direction}, uid => $o->{uid} });
	} else {
		# can't go that way
		$kernel->post($sender->ID => send => "There is no exit that way");
	}

	return 1;
}

sub redef_count {
	my $ref = shift;
	sub {
		if ($_[0] =~ m/[Ss]ubroutine ([\w:]+) redefined/) {
			my $su = $1;
			$$ref++;
			local($|) = 1;
			return;
		}
		warn @_;
	};
}
																									
1;

