package time;

use strict;
use POE;

sub new {
	my $class = shift;
	
	# every module must have a 'provides' arrayhash ref
	# and a help arrayref in its obj
	my $self = bless {
		# used to match events to commands, keep the # down to very few
		# this is matched against every line that starts with ! or . or /
		# m/^[\!|\.|\/]/
		provides => {
			timezone_help => 'm/^\!timezone$/i',
			time_help => 'm/^\.time$/i',
			
			time_now => 'm/^\.time now/i',
			timezone => 'm/^\!timezone (.*)/i',
		},
		# used to fill the help list (optional, but very helpful :))
		# format is '.command' or '!command' or '/command'
		help => [
			'.time',
			'!timezone',
		],
		
		# internals
		defaults => {
			timezone => 'PDT',
		},
	},$class;

	my $session = POE::Session->create(
		heap => {
			module_name => 'Time',
		},
		#options => { default=>1, trace=>1 },
		heap => shift,
		object_states => [
			$self => {
				_start			=> '_start',
				_stop			=> '_stop',
				shutdown		=> 'shutdown',
				
				time_help		=> 'time_help',
				timezone_help	=> 'timezone_help',
				
				time_now		=> 'time_now',
				timezone		=> 'timezone',

				say_time		=> 'say_time',

				# provided for us
				logged_in		=> 'logged_in',
				logged_out		=> 'logged_out',
			}
		],
	);

	$self->{sid} = $session->ID;

	return $self;
}


sub _start {
	my ($kernel, $heap, $session, $sender, $self) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT];

	$heap->{c} = $sender->ID;
	
	# keep us alive
	
    $kernel->refcount_increment($session->ID, "module::".__PACKAGE__);
}

sub _stop {
	my ($kernel, $heap, $session, $sender, $self) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT];

	
}

sub shutdown {
	my ($kernel, $heap, $session, $sender, $self) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT];

	$kernel->alarm_remove_all();
    $kernel->refcount_decrement($session->ID, "module::".__PACKAGE__);

	return $session->ID;
}

sub logged_in {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	$heap->{user_sids}{$o->{uid}} = $sender->ID;
	
	$kernel->delay_set(say_time => 1 => $o);
}

sub logged_out {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	delete $heap->{user_sids}{$o->{uid}};
}

sub time_help {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

	$kernel->post($sender->ID => send => "Time help",".time now\t\tPrints current time/date");
}

sub timezone_help {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

}

sub time_now {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];
	# sender is the control session the user is logged into
	# use my $cheap = $sender->get_heap(); to access the control
	# sessions' heap.  Be carefull with what you mess with...
	
	$kernel->post($sender->ID => send => "The time is ".localtime());
}

sub timezone {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];

}

sub say_time {
	my ($kernel, $heap, $session, $sender, $self, $o) = @_[KERNEL, HEAP, SESSION, SENDER, OBJECT, ARG0];
	
	return unless (exists ($heap->{user_sids}{$o->{uid}}));

	$kernel->post($heap->{user_sids}{$o->{uid}} => send => "\0[bold;cyan]It is now ".localtime()."\0[reset]");
	$kernel->delay_set(say_time => (60*60) => $o);
}

1;
